@@ -26,11 +26,11 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <gio/gfiledescriptorbased.h>
+#include "libglnx.h"
 #include "ostree.h"
 #include "ostree-core-private.h"
 #include "ostree-chain-input-stream.h"
 #include "otutil.h"
-#include "libgsystem.h"
 
 #define ALIGN_VALUE(this, boundary) \
   (( ((unsigned long)(this)) + (((unsigned long)(boundary)) -1)) & (~(((unsigned long)(boundary))-1)))
@@ -47,7 +47,7 @@ zlib_file_header_parse (GVariant         *metadata,
                         GError          **error);
 
 /**
- * SECTION:libostree-core
+ * SECTION:ostree-core
  * @title: Core repository-independent functions
  * @short_description: Create, validate, and convert core data types
  *
@@ -123,7 +123,7 @@ ostree_parse_refspec (const char   *refspec,
 {
   gboolean ret = FALSE;
   GMatchInfo *match = NULL;
-  char *remote;
+  g_autofree char *remote = NULL;
 
   static gsize regex_initialized;
   static GRegex *regex;
@@ -154,8 +154,11 @@ ostree_parse_refspec (const char   *refspec,
     }
 
   ret = TRUE;
-  *out_remote = remote;
-  *out_ref = g_match_info_fetch (match, 2);
+
+  if (out_remote)
+    *out_remote = g_steal_pointer (&remote);
+  if (out_ref != NULL)
+    *out_ref = g_match_info_fetch (match, 2);
  out:
   if (match)
     g_match_info_unref (match);
@@ -200,22 +203,20 @@ ostree_validate_rev (const char *rev,
   return ret;
 }
 
-static GVariant *
-file_header_new (GFileInfo         *file_info,
-                 GVariant          *xattrs)
+GVariant *
+_ostree_file_header_new (GFileInfo         *file_info,
+                         GVariant          *xattrs)
 {
   guint32 uid;
   guint32 gid;
   guint32 mode;
-  guint32 rdev;
   const char *symlink_target;
   GVariant *ret;
-  gs_unref_variant GVariant *tmp_xattrs = NULL;
+  g_autoptr(GVariant) tmp_xattrs = NULL;
 
   uid = g_file_info_get_attribute_uint32 (file_info, "unix::uid");
   gid = g_file_info_get_attribute_uint32 (file_info, "unix::gid");
   mode = g_file_info_get_attribute_uint32 (file_info, "unix::mode");
-  rdev = g_file_info_get_attribute_uint32 (file_info, "unix::rdev");
 
   if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_SYMBOLIC_LINK)
     symlink_target = g_file_info_get_symlink_target (file_info);
@@ -226,7 +227,7 @@ file_header_new (GFileInfo         *file_info,
     tmp_xattrs = g_variant_ref_sink (g_variant_new_array (G_VARIANT_TYPE ("(ayay)"), NULL, 0));
 
   ret = g_variant_new ("(uuuus@a(ayay))", GUINT32_TO_BE (uid),
-                       GUINT32_TO_BE (gid), GUINT32_TO_BE (mode), GUINT32_TO_BE (rdev),
+                       GUINT32_TO_BE (gid), GUINT32_TO_BE (mode), 0,
                        symlink_target, xattrs ? xattrs : tmp_xattrs);
   g_variant_ref_sink (ret);
   return ret;
@@ -247,16 +248,14 @@ _ostree_zlib_file_header_new (GFileInfo         *file_info,
   guint32 uid;
   guint32 gid;
   guint32 mode;
-  guint32 rdev;
   const char *symlink_target;
   GVariant *ret;
-  gs_unref_variant GVariant *tmp_xattrs = NULL;
+  g_autoptr(GVariant) tmp_xattrs = NULL;
 
   size = g_file_info_get_size (file_info);
   uid = g_file_info_get_attribute_uint32 (file_info, "unix::uid");
   gid = g_file_info_get_attribute_uint32 (file_info, "unix::gid");
   mode = g_file_info_get_attribute_uint32 (file_info, "unix::mode");
-  rdev = g_file_info_get_attribute_uint32 (file_info, "unix::rdev");
 
   if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_SYMBOLIC_LINK)
     symlink_target = g_file_info_get_symlink_target (file_info);
@@ -268,7 +267,7 @@ _ostree_zlib_file_header_new (GFileInfo         *file_info,
 
   ret = g_variant_new ("(tuuuus@a(ayay))",
                        GUINT64_TO_BE (size), GUINT32_TO_BE (uid),
-                       GUINT32_TO_BE (gid), GUINT32_TO_BE (mode), GUINT32_TO_BE (rdev),
+                       GUINT32_TO_BE (gid), GUINT32_TO_BE (mode), 0,
                        symlink_target, xattrs ? xattrs : tmp_xattrs);
   g_variant_ref_sink (ret);
   return ret;
@@ -401,6 +400,98 @@ write_file_header_update_checksum (GOutputStream         *out,
   return ret;
 }
 
+/*
+ * header_and_input_to_stream:
+ * @file_header: A file header
+ * @input: File raw content stream
+ * @out_input: (out): Serialized object stream
+ * @out_header_size: (out): Length of the header
+ * @cancellable: Cancellable
+ * @error: Error
+ *
+ * Combines @file_header and @input into a single stream.
+ */
+static gboolean
+header_and_input_to_stream (GVariant           *file_header,
+                            GInputStream       *input,
+                            GInputStream      **out_input,
+                            guint64            *out_header_size,
+                            GCancellable       *cancellable,
+                            GError            **error)
+{
+  gpointer header_data;
+  gsize header_size;
+  g_autoptr(GInputStream) ret_input = NULL;
+  g_autoptr(GPtrArray) streams = NULL;
+  g_autoptr(GOutputStream) header_out_stream = NULL;
+  g_autoptr(GInputStream) header_in_stream = NULL;
+
+  header_out_stream = g_memory_output_stream_new (NULL, 0, g_realloc, g_free);
+
+  if (!_ostree_write_variant_with_size (header_out_stream, file_header, 0, NULL, NULL,
+                                        cancellable, error))
+    return FALSE;
+
+  if (!g_output_stream_close (header_out_stream, cancellable, error))
+    return FALSE;
+
+  header_size = g_memory_output_stream_get_data_size ((GMemoryOutputStream*) header_out_stream);
+  header_data = g_memory_output_stream_steal_data ((GMemoryOutputStream*) header_out_stream);
+  header_in_stream = g_memory_input_stream_new_from_data (header_data, header_size, g_free);
+
+  streams = g_ptr_array_new_with_free_func ((GDestroyNotify)g_object_unref);
+
+  g_ptr_array_add (streams, g_object_ref (header_in_stream));
+  if (input)
+    g_ptr_array_add (streams, g_object_ref (input));
+
+  ret_input = (GInputStream*)ostree_chain_input_stream_new (streams);
+  ot_transfer_out_value (out_input, &ret_input);
+  if (out_header_size)
+    *out_header_size = header_size;
+
+  return TRUE;
+}
+
+/**
+ * ostree_raw_file_to_archive_z2_stream:
+ * @input: File raw content stream
+ * @file_info: A file info
+ * @xattrs: (allow-none): Optional extended attributes
+ * @out_input: (out): Serialized object stream
+ * @cancellable: Cancellable
+ * @error: Error
+ *
+ * Convert from a "bare" file representation into an
+ * OSTREE_OBJECT_TYPE_FILE stream suitable for ostree pull.
+ */
+gboolean
+ostree_raw_file_to_archive_z2_stream (GInputStream       *input,
+                                      GFileInfo          *file_info,
+                                      GVariant           *xattrs,
+                                      GInputStream      **out_input,
+                                      GCancellable       *cancellable,
+                                      GError            **error)
+{
+  g_autoptr(GVariant) file_header = NULL;
+  g_autoptr(GInputStream) zlib_input = NULL;
+
+  file_header = _ostree_zlib_file_header_new (file_info, xattrs);
+  if (input != NULL)
+    {
+      g_autoptr(GConverter) zlib_compressor = NULL;
+
+      zlib_compressor = G_CONVERTER (g_zlib_compressor_new (G_ZLIB_COMPRESSOR_FORMAT_RAW, 9));
+      zlib_input = g_converter_input_stream_new (input, zlib_compressor);
+    }
+  return header_and_input_to_stream (file_header,
+                                     zlib_input,
+                                     out_input,
+                                     NULL,
+                                     cancellable,
+                                     error);
+}
+
 /**
  * ostree_raw_file_to_content_stream:
  * @input: File raw content stream
@@ -424,44 +515,20 @@ ostree_raw_file_to_content_stream (GInputStream       *input,
                                    GCancellable       *cancellable,
                                    GError            **error)
 {
-  gboolean ret = FALSE;
-  gpointer header_data;
-  gsize header_size;
-  gs_unref_object GInputStream *ret_input = NULL;
-  gs_unref_variant GVariant *file_header = NULL;
-  gs_unref_ptrarray GPtrArray *streams = NULL;
-  gs_unref_object GOutputStream *header_out_stream = NULL;
-  gs_unref_object GInputStream *header_in_stream = NULL;
+  g_autoptr(GVariant) file_header = NULL;
+  guint64 header_size;
 
-  file_header = file_header_new (file_info, xattrs);
-
-  header_out_stream = g_memory_output_stream_new (NULL, 0, g_realloc, g_free);
-
-  if (!_ostree_write_variant_with_size (header_out_stream, file_header, 0, NULL, NULL,
-                                        cancellable, error))
-    goto out;
-
-  if (!g_output_stream_close (header_out_stream, cancellable, error))
-    goto out;
-
-  header_size = g_memory_output_stream_get_data_size ((GMemoryOutputStream*) header_out_stream);
-  header_data = g_memory_output_stream_steal_data ((GMemoryOutputStream*) header_out_stream);
-  header_in_stream = g_memory_input_stream_new_from_data (header_data, header_size, g_free);
-
-  streams = g_ptr_array_new_with_free_func ((GDestroyNotify)g_object_unref);
-
-  g_ptr_array_add (streams, g_object_ref (header_in_stream));
-  if (input)
-    g_ptr_array_add (streams, g_object_ref (input));
-  
-  ret_input = (GInputStream*)ostree_chain_input_stream_new (streams);
-
-  ret = TRUE;
-  ot_transfer_out_value (out_input, &ret_input);
+  file_header = _ostree_file_header_new (file_info, xattrs);
+  if (!header_and_input_to_stream (file_header,
+                                   input,
+                                   out_input,
+                                   &header_size,
+                                   cancellable,
+                                   error))
+    return FALSE;
   if (out_length)
     *out_length = header_size + g_file_info_get_size (file_info);
- out:
-  return ret;
+  return TRUE;
 }
 
 /**
@@ -494,11 +561,11 @@ ostree_content_stream_parse (gboolean                compressed,
   guint32 archive_header_size;
   guchar dummy[4];
   gsize bytes_read;
-  gs_unref_object GInputStream *ret_input = NULL;
-  gs_unref_object GFileInfo *ret_file_info = NULL;
-  gs_unref_variant GVariant *ret_xattrs = NULL;
-  gs_unref_variant GVariant *file_header = NULL;
-  gs_free guchar *buf = NULL;
+  g_autoptr(GInputStream) ret_input = NULL;
+  g_autoptr(GFileInfo) ret_file_info = NULL;
+  g_autoptr(GVariant) ret_xattrs = NULL;
+  g_autoptr(GVariant) file_header = NULL;
+  g_autofree guchar *buf = NULL;
 
   if (!g_input_stream_read_all (input,
                                 &archive_header_size, 4, &bytes_read,
@@ -562,7 +629,7 @@ ostree_content_stream_parse (gboolean                compressed,
        **/
       if (compressed)
         {
-          gs_unref_object GConverter *zlib_decomp = (GConverter*)g_zlib_decompressor_new (G_ZLIB_COMPRESSOR_FORMAT_RAW);
+          g_autoptr(GConverter) zlib_decomp = (GConverter*)g_zlib_decompressor_new (G_ZLIB_COMPRESSOR_FORMAT_RAW);
           ret_input = g_converter_input_stream_new (input, zlib_decomp);
         }
       else
@@ -577,6 +644,60 @@ ostree_content_stream_parse (gboolean                compressed,
   return ret;
 }
 
+/**
+ * ostree_content_file_parse_at:
+ * @compressed: Whether or not the stream is zlib-compressed
+ * @parent_dfd: Directory file descriptor
+ * @path: Subpath
+ * @trusted: If %TRUE, assume the content has been validated
+ * @out_input: (out): The raw file content stream
+ * @out_file_info: (out): Normal metadata 
+ * @out_xattrs: (out): Extended attributes
+ * @cancellable: Cancellable
+ * @error: Error
+ *
+ * A thin wrapper for ostree_content_stream_parse(); this function
+ * converts an object content stream back into components.
+ */
+gboolean
+ostree_content_file_parse_at (gboolean                compressed,
+                              int                     parent_dfd,
+                              const char             *path,
+                              gboolean                trusted,
+                              GInputStream          **out_input,
+                              GFileInfo             **out_file_info,
+                              GVariant              **out_xattrs,
+                              GCancellable           *cancellable,
+                              GError                **error)
+{
+  gboolean ret = FALSE;
+  struct stat stbuf;
+  g_autoptr(GInputStream) file_input = NULL;
+  g_autoptr(GInputStream) ret_input = NULL;
+  g_autoptr(GFileInfo) ret_file_info = NULL;
+  g_autoptr(GVariant) ret_xattrs = NULL;
+
+  if (!ot_openat_read_stream (parent_dfd, path, TRUE, &file_input,
+                              cancellable, error))
+    goto out;
+      
+  if (!glnx_stream_fstat ((GFileDescriptorBased*)file_input, &stbuf, error))
+    goto out;
+  
+  if (!ostree_content_stream_parse (compressed, file_input, stbuf.st_size, trusted,
+                                    out_input ? &ret_input : NULL,
+                                    &ret_file_info, &ret_xattrs,
+                                    cancellable, error))
+    goto out;
+      
+  ret = TRUE;
+  ot_transfer_out_value (out_input, &ret_input);
+  ot_transfer_out_value (out_file_info, &ret_file_info);
+  ot_transfer_out_value (out_xattrs, &ret_xattrs);
+ out:
+  return ret;
+}
+
 /**
  * ostree_content_file_parse:
  * @compressed: Whether or not the stream is zlib-compressed
@@ -601,53 +722,11 @@ ostree_content_file_parse (gboolean                compressed,
                            GCancellable           *cancellable,
                            GError                **error)
 {
-  gboolean ret = FALSE;
-  guint64 length;
-  struct stat stbuf;
-  gs_unref_object GInputStream *file_input = NULL;
-  gs_unref_object GInputStream *ret_input = NULL;
-  gs_unref_object GFileInfo *ret_file_info = NULL;
-  gs_unref_variant GVariant *ret_xattrs = NULL;
-
-  if (out_input)
-    {
-      file_input = (GInputStream*)gs_file_read_noatime (content_path, cancellable, error);
-      if (!file_input)
-        goto out;
-      
-      if (!gs_stream_fstat ((GFileDescriptorBased*)file_input, &stbuf, cancellable, error))
-        goto out;
-
-      length = stbuf.st_size;
-    }
-  else
-    {
-      GMappedFile *mmaped;
-      GBytes *bytes;
-
-      mmaped = gs_file_map_noatime (content_path, cancellable, error);
-      if (!mmaped)
-        goto out;
-      bytes = g_mapped_file_get_bytes (mmaped);
-      g_mapped_file_unref (mmaped);
-      mmaped = NULL;
-      file_input = g_memory_input_stream_new_from_bytes (bytes);
-      length = g_bytes_get_size (bytes);
-      g_bytes_unref (bytes);
-    }
-
-  if (!ostree_content_stream_parse (compressed, file_input, length, trusted,
-                                    out_input ? &ret_input : NULL,
-                                    &ret_file_info, &ret_xattrs,
-                                    cancellable, error))
-    goto out;
-      
-  ret = TRUE;
-  ot_transfer_out_value (out_input, &ret_input);
-  ot_transfer_out_value (out_file_info, &ret_file_info);
-  ot_transfer_out_value (out_xattrs, &ret_xattrs);
- out:
-  return ret;
+  return ostree_content_file_parse_at (compressed, AT_FDCWD,
+                                       gs_file_get_path_cached (content_path),
+                                       trusted,
+                                       out_input, out_file_info, out_xattrs,
+                                       cancellable, error);
 }
 
 /**
@@ -672,7 +751,7 @@ ostree_checksum_file_from_input (GFileInfo        *file_info,
                                  GError          **error)
 {
   gboolean ret = FALSE;
-  gs_free guchar *ret_csum = NULL;
+  g_autofree guchar *ret_csum = NULL;
   GChecksum *checksum = NULL;
 
   checksum = g_checksum_new (G_CHECKSUM_SHA256);
@@ -684,16 +763,16 @@ ostree_checksum_file_from_input (GFileInfo        *file_info,
     }
   else if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_DIRECTORY)
     {
-      gs_unref_variant GVariant *dirmeta = ostree_create_directory_metadata (file_info, xattrs);
+      g_autoptr(GVariant) dirmeta = ostree_create_directory_metadata (file_info, xattrs);
       g_checksum_update (checksum, g_variant_get_data (dirmeta),
                          g_variant_get_size (dirmeta));
       
     }
   else
     {
-      gs_unref_variant GVariant *file_header = NULL;
+      g_autoptr(GVariant) file_header = NULL;
 
-      file_header = file_header_new (file_info, xattrs);
+      file_header = _ostree_file_header_new (file_info, xattrs);
 
       if (!write_file_header_update_checksum (NULL, file_header, checksum,
                                               cancellable, error))
@@ -733,10 +812,10 @@ ostree_checksum_file (GFile            *f,
                       GError          **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object GFileInfo *file_info = NULL;
-  gs_unref_object GInputStream *in = NULL;
-  gs_unref_variant GVariant *xattrs = NULL;
-  gs_free guchar *ret_csum = NULL;
+  g_autoptr(GFileInfo) file_info = NULL;
+  g_autoptr(GInputStream) in = NULL;
+  g_autoptr(GVariant) xattrs = NULL;
+  g_autofree guchar *ret_csum = NULL;
 
   if (g_cancellable_set_error_if_cancelled (cancellable, error))
     return FALSE;
@@ -756,7 +835,8 @@ ostree_checksum_file (GFile            *f,
 
   if (objtype == OSTREE_OBJECT_TYPE_FILE)
     {
-      if (!gs_file_get_all_xattrs (f, &xattrs, cancellable, error))
+      if (!glnx_dfd_name_get_all_xattrs (AT_FDCWD, gs_file_get_path_cached (f),
+                                         &xattrs, cancellable, error))
         goto out;
     }
 
@@ -909,14 +989,13 @@ _ostree_make_temporary_symlink_at (int             tmp_dirfd,
                                    GError        **error)
 {
   gboolean ret = FALSE;
-  gs_free char *tmpname = NULL;
+  char *tmpname = g_strdup ("tmplink.XXXXXX");
   guint i;
   const int max_attempts = 128;
 
   for (i = 0; i < max_attempts; i++)
     {
-      g_free (tmpname);
-      tmpname = gs_fileutil_gen_tmp_name (NULL, NULL);
+      glnx_gen_temp_name (tmpname);
       if (symlinkat (target, tmp_dirfd, tmpname) < 0)
         {
           if (errno == EEXIST)
@@ -940,7 +1019,8 @@ _ostree_make_temporary_symlink_at (int             tmp_dirfd,
     }
 
   ret = TRUE;
-  gs_transfer_out_value (out_name, &tmpname);
+  if (out_name)
+    *out_name = g_steal_pointer (&tmpname);
  out:
   return ret;
 }
@@ -965,6 +1045,10 @@ ostree_object_type_to_string (OstreeObjectType objtype)
       return "dirmeta";
     case OSTREE_OBJECT_TYPE_COMMIT:
       return "commit";
+    case OSTREE_OBJECT_TYPE_TOMBSTONE_COMMIT:
+      return "tombstone-commit";
+    case OSTREE_OBJECT_TYPE_COMMIT_META:
+      return "commitmeta";
     default:
       g_assert_not_reached ();
       return NULL;
@@ -988,6 +1072,10 @@ ostree_object_type_from_string (const char *str)
     return OSTREE_OBJECT_TYPE_DIR_META;
   else if (!strcmp (str, "commit"))
     return OSTREE_OBJECT_TYPE_COMMIT;
+  else if (!strcmp (str, "tombstone-commit"))
+    return OSTREE_OBJECT_TYPE_TOMBSTONE_COMMIT;
+  else if (!strcmp (str, "commitmeta"))
+    return OSTREE_OBJECT_TYPE_COMMIT_META;
   g_assert_not_reached ();
   return 0;
 }
@@ -1057,7 +1145,7 @@ int
 ostree_cmp_checksum_bytes (const guchar *a,
                            const guchar *b)
 {
-  return memcmp (a, b, 32);
+  return memcmp (a, b, OSTREE_SHA256_DIGEST_LEN);
 }
 
 /**
@@ -1072,7 +1160,7 @@ ostree_object_name_serialize (const char *checksum,
                               OstreeObjectType objtype)
 {
   g_assert (objtype >= OSTREE_OBJECT_TYPE_FILE
-            && objtype <= OSTREE_OBJECT_TYPE_COMMIT);
+            && objtype <= OSTREE_OBJECT_TYPE_LAST);
   return g_variant_new ("(su)", checksum, (guint32)objtype);
 }
 
@@ -1095,6 +1183,35 @@ ostree_object_name_deserialize (GVariant         *variant,
   *out_objtype = (OstreeObjectType)objtype_u32;
 }
 
+/**
+ * ostree_checksum_b64_inplace_to_bytes: (skip)
+ * @checksum: (array fixed-size=32): An binary checksum of length 32
+ * @buf: Output location, must be at least 45 bytes in length
+ *
+ * Overwrite the contents of @buf with stringified version of @csum.
+ */
+void
+ostree_checksum_b64_inplace_to_bytes (const char *checksum,
+                                      guchar     *buf)
+{
+  int state = 0;
+  guint save = 0;
+  char tmpbuf[44];
+  int i;
+
+  for (i = 0; i < 43; i++)
+    {
+      char c = checksum[i];
+      if (c == '_')
+        tmpbuf[i] = '/';
+      else
+        tmpbuf[i] = c;
+    }
+  tmpbuf[43] = '=';
+
+  g_base64_decode_step (tmpbuf, sizeof (tmpbuf), (guchar *) buf, &state, &save);
+}
+
 /**
  * ostree_checksum_inplace_to_bytes:
  * @checksum: a SHA256 string
@@ -1110,7 +1227,7 @@ ostree_checksum_inplace_to_bytes (const char *checksum,
   guint i;
   guint j;
 
-  for (i = 0, j = 0; i < 32; i += 1, j += 2)
+  for (i = 0, j = 0; i < OSTREE_SHA256_DIGEST_LEN; i += 1, j += 2)
     {
       gint big, little;
 
@@ -1136,7 +1253,7 @@ ostree_checksum_inplace_to_bytes (const char *checksum,
 guchar *
 ostree_checksum_to_bytes (const char *checksum)
 {
-  guchar *ret = g_malloc (32);
+  guchar *ret = g_malloc (OSTREE_SHA256_DIGEST_LEN);
   ostree_checksum_inplace_to_bytes (checksum, ret);
   return ret;
 }
@@ -1150,15 +1267,29 @@ ostree_checksum_to_bytes (const char *checksum)
 GVariant *
 ostree_checksum_to_bytes_v (const char *checksum)
 {
-  guchar result[32];
+  guchar result[OSTREE_SHA256_DIGEST_LEN];
   ostree_checksum_inplace_to_bytes (checksum, result);
-  return ot_gvariant_new_bytearray ((guchar*)result, 32);
+  return ot_gvariant_new_bytearray ((guchar*)result, OSTREE_SHA256_DIGEST_LEN);
+}
+
+/**
+ * ostree_checksum_b64_to_bytes:
+ * @checksum: An ASCII checksum
+ *
+ * Returns: (transfer full) (array fixed-size=32): Binary version of @checksum.
+ */
+guchar *
+ostree_checksum_b64_to_bytes (const char *checksum)
+{
+  guchar *ret = g_malloc (32);
+  ostree_checksum_b64_inplace_to_bytes (checksum, ret);
+  return ret;
 }
 
 /**
  * ostree_checksum_inplace_from_bytes: (skip)
  * @csum: (array fixed-size=32): An binary checksum of length 32
- * @buf: Output location, must be at least 65 bytes in length
+ * @buf: Output location, must be at least OSTREE_SHA256_STRING_LEN+1 bytes in length
  *
  * Overwrite the contents of @buf with stringified version of @csum.
  */
@@ -1169,7 +1300,7 @@ ostree_checksum_inplace_from_bytes (const guchar *csum,
   static const gchar hexchars[] = "0123456789abcdef";
   guint i, j;
 
-  for (i = 0, j = 0; i < 32; i++, j += 2)
+  for (i = 0, j = 0; i < OSTREE_SHA256_DIGEST_LEN; i++, j += 2)
     {
       guchar byte = csum[i];
       buf[j] = hexchars[byte >> 4];
@@ -1178,6 +1309,48 @@ ostree_checksum_inplace_from_bytes (const guchar *csum,
   buf[j] = '\0';
 }
 
+/**
+ * ostree_checksum_b64_inplace_from_bytes: (skip)
+ * @csum: (array fixed-size=32): An binary checksum of length 32
+ * @buf: Output location, must be at least 44 bytes in length
+ *
+ * Overwrite the contents of @buf with modified base64 encoding of @csum.
+ * The "modified" term refers to the fact that instead of '/', the '_'
+ * character is used.
+ */
+void
+ostree_checksum_b64_inplace_from_bytes (const guchar *csum,
+                                        char         *buf)
+{
+  char tmpbuf[44];
+  int save = 0;
+  int state = 0;
+  gsize outlen;
+  int i;
+
+  /* At some point, we can optimize this, but for now it's
+   * a lot easier to reuse GLib's base64 encoder and postprocess it
+   * to replace the '/' with '_'.
+   */
+  outlen = g_base64_encode_step (csum, OSTREE_SHA256_DIGEST_LEN, FALSE, tmpbuf, &state, &save);
+  outlen += g_base64_encode_close (FALSE, tmpbuf+outlen, &state, &save);
+  g_assert (outlen == 44);
+
+  for (i = 0; i < sizeof (tmpbuf); i++)
+    {
+      char c = tmpbuf[i];
+      if (c == '=')
+        {
+          g_assert (i == 43);
+          buf[i] = '\0';
+        }
+      else if (c == '/')
+        buf[i] = '_';
+      else
+        buf[i] = c;
+    }
+}
+
 /**
  * ostree_checksum_from_bytes:
  * @csum: (array fixed-size=32): An binary checksum of length 32
@@ -1187,7 +1360,7 @@ ostree_checksum_inplace_from_bytes (const guchar *csum,
 char *
 ostree_checksum_from_bytes (const guchar *csum)
 {
-  char *ret = g_malloc (65);
+  char *ret = g_malloc (OSTREE_SHA256_STRING_LEN+1);
   ostree_checksum_inplace_from_bytes (csum, ret);
   return ret;
 }
@@ -1204,6 +1377,23 @@ ostree_checksum_from_bytes_v (GVariant *csum_v)
   return ostree_checksum_from_bytes (ostree_checksum_bytes_peek (csum_v));
 }
 
+/**
+ * ostree_checksum_b64_from_bytes:
+ * @csum: (array fixed-size=32): An binary checksum of length 32
+ *
+ * Returns: (transfer full): Modified base64 encoding of @csum
+ *
+ * The "modified" term refers to the fact that instead of '/', the '_'
+ * character is used.
+ */
+char *
+ostree_checksum_b64_from_bytes (const guchar *csum)
+{
+  char *ret = g_malloc (44);
+  ostree_checksum_b64_inplace_from_bytes (csum, ret);
+  return ret;
+}
+
 /**
  * ostree_checksum_bytes_peek:
  * @bytes: #GVariant of type ay
@@ -1216,7 +1406,7 @@ ostree_checksum_bytes_peek (GVariant *bytes)
   gsize n_elts;
   const guchar *ret;
   ret = g_variant_get_fixed_array (bytes, &n_elts, 1);
-  if (G_UNLIKELY (n_elts != 32))
+  if (G_UNLIKELY (n_elts != OSTREE_SHA256_DIGEST_LEN))
     return NULL;
   return ret;
 }
@@ -1260,35 +1450,39 @@ _ostree_loose_path (char              *buf,
                     const char        *checksum,
                     OstreeObjectType   objtype,
                     OstreeRepoMode     mode)
-{
-  _ostree_loose_path_with_suffix (buf, checksum, objtype, mode, "");
-}
-
-/*
- * _ostree_loose_path_with_suffix:
- * @buf: Output buffer, must be _OSTREE_LOOSE_PATH_MAX in size
- * @checksum: ASCII checksum
- * @objtype: Object type
- * @mode: Repository mode
- *
- * Like _ostree_loose_path, but also append a further arbitrary
- * suffix; useful for finding non-core objects.
- */
-void
-_ostree_loose_path_with_suffix (char              *buf,
-                                const char        *checksum,
-                                OstreeObjectType   objtype,
-                                OstreeRepoMode     mode,
-                                const char        *suffix)
 {
   *buf = checksum[0];
   buf++;
   *buf = checksum[1];
   buf++;
-  snprintf (buf, _OSTREE_LOOSE_PATH_MAX - 2, "/%s.%s%s%s",
+  snprintf (buf, _OSTREE_LOOSE_PATH_MAX - 2, "/%s.%s%s",
             checksum + 2, ostree_object_type_to_string (objtype),
-            (!OSTREE_OBJECT_TYPE_IS_META (objtype) && mode == OSTREE_REPO_MODE_ARCHIVE_Z2) ? "z" : "",
-            suffix);
+            (!OSTREE_OBJECT_TYPE_IS_META (objtype) && mode == OSTREE_REPO_MODE_ARCHIVE_Z2) ? "z" : "");
+}
+
+/**
+ * _ostree_header_gfile_info_new:
+ * @mode: File mode
+ * @uid: File uid
+ * @gid: File gid
+ *
+ * OSTree only stores a subset of file attributes; for example,
+ * timestamps are intentionally not stored.  This function creates a
+ * #GFileInfo based on the attributes of a `struct stat` that match
+ * those file attributes.
+ *
+ * Returns: (transfer full): A new #GFileInfo mapping a subset of @stbuf.
+ */
+GFileInfo *
+_ostree_header_gfile_info_new (mode_t mode, uid_t uid, gid_t gid)
+{
+  GFileInfo *ret = g_file_info_new ();
+  g_file_info_set_attribute_uint32 (ret, "standard::type", ot_gfile_type_for_mode (mode));
+  g_file_info_set_attribute_boolean (ret, "standard::is-symlink", S_ISLNK (mode));
+  g_file_info_set_attribute_uint32 (ret, "unix::uid", uid);
+  g_file_info_set_attribute_uint32 (ret, "unix::gid", gid);
+  g_file_info_set_attribute_uint32 (ret, "unix::mode", mode);
+  return ret;
 }
 
 /*
@@ -1306,7 +1500,7 @@ _ostree_get_relative_object_path (const char         *checksum,
 {
   GString *path;
 
-  g_assert (strlen (checksum) == 64);
+  g_assert (strlen (checksum) == OSTREE_SHA256_STRING_LEN);
 
   path = g_string_new ("objects/");
 
@@ -1322,10 +1516,63 @@ _ostree_get_relative_object_path (const char         *checksum,
 }
 
 char *
-_ostree_get_relative_static_delta_path (const char        *from,
-                                        const char        *to)
+_ostree_get_relative_static_delta_path (const char *from,
+                                        const char *to,
+                                        const char *target)
 {
-  return g_strdup_printf ("deltas/%s-%s/meta", from, to);
+  guint8 csum_to[OSTREE_SHA256_DIGEST_LEN];
+  char to_b64[44];
+  guint8 csum_to_copy[OSTREE_SHA256_DIGEST_LEN];
+  GString *ret = g_string_new ("deltas/");
+
+  ostree_checksum_inplace_to_bytes (to, csum_to);
+  ostree_checksum_b64_inplace_from_bytes (csum_to, to_b64);
+  ostree_checksum_b64_inplace_to_bytes (to_b64, csum_to_copy);
+
+  g_assert (memcmp (csum_to, csum_to_copy, OSTREE_SHA256_DIGEST_LEN) == 0);
+
+  if (from != NULL)
+    {
+      guint8 csum_from[OSTREE_SHA256_DIGEST_LEN];
+      char from_b64[44];
+
+      ostree_checksum_inplace_to_bytes (from, csum_from);
+      ostree_checksum_b64_inplace_from_bytes (csum_from, from_b64);
+
+      g_string_append_c (ret, from_b64[0]);
+      g_string_append_c (ret, from_b64[1]);
+      g_string_append_c (ret, '/');
+      g_string_append (ret, from_b64 + 2);
+      g_string_append_c (ret, '-');
+    }
+
+  g_string_append_c (ret, to_b64[0]);
+  g_string_append_c (ret, to_b64[1]);
+  if (from == NULL)
+    g_string_append_c (ret, '/');
+  g_string_append (ret, to_b64 + 2);
+
+  if (target != NULL)
+    {
+      g_string_append_c (ret, '/');
+      g_string_append (ret, target);
+    }
+  
+  return g_string_free (ret, FALSE);
+}
+
+char *
+_ostree_get_relative_static_delta_superblock_path (const char        *from,
+                                                   const char        *to)
+{
+  return _ostree_get_relative_static_delta_path (from, to, "superblock");
+}
+
+char *
+_ostree_get_relative_static_delta_detachedmeta_path (const char        *from,
+                                                     const char        *to)
+{
+  return _ostree_get_relative_static_delta_path (from, to, "meta");
 }
 
 char *
@@ -1333,7 +1580,27 @@ _ostree_get_relative_static_delta_part_path (const char        *from,
                                              const char        *to,
                                              guint              i)
 {
-  return g_strdup_printf ("deltas/%s-%s/%u", from, to, i);
+  g_autofree char *partstr = g_strdup_printf ("%u", i);
+  return _ostree_get_relative_static_delta_path (from, to, partstr);
+}
+
+void
+_ostree_parse_delta_name (const char  *delta_name,
+                          char        **out_from,
+                          char        **out_to)
+{
+  g_auto(GStrv) parts = g_strsplit (delta_name, "-", 2);
+
+  *out_from = *out_to = NULL;
+  if (parts[0] && parts[1])
+    {
+      ot_transfer_out_value (out_from, &parts[0]);
+      ot_transfer_out_value (out_to, &parts[1]);
+    }
+  else
+    {
+      ot_transfer_out_value (out_to, &parts[0]);
+    }
 }
 
 /*
@@ -1355,24 +1622,24 @@ file_header_parse (GVariant         *metadata,
   gboolean ret = FALSE;
   guint32 uid, gid, mode, rdev;
   const char *symlink_target;
-  gs_unref_object GFileInfo *ret_file_info = NULL;
-  gs_unref_variant GVariant *ret_xattrs = NULL;
+  g_autoptr(GFileInfo) ret_file_info = NULL;
+  g_autoptr(GVariant) ret_xattrs = NULL;
 
   g_variant_get (metadata, "(uuuu&s@a(ayay))",
                  &uid, &gid, &mode, &rdev,
                  &symlink_target, &ret_xattrs);
+  if (rdev != 0)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                   "Corrupted archive file; invalid rdev %u", GUINT32_FROM_BE (rdev));
+      goto out;
+    }
 
   uid = GUINT32_FROM_BE (uid);
   gid = GUINT32_FROM_BE (gid);
   mode = GUINT32_FROM_BE (mode);
-  rdev = GUINT32_FROM_BE (rdev);
 
-  ret_file_info = g_file_info_new ();
-  g_file_info_set_attribute_uint32 (ret_file_info, "standard::type", ot_gfile_type_for_mode (mode));
-  g_file_info_set_attribute_boolean (ret_file_info, "standard::is-symlink", S_ISLNK (mode));
-  g_file_info_set_attribute_uint32 (ret_file_info, "unix::uid", uid);
-  g_file_info_set_attribute_uint32 (ret_file_info, "unix::gid", gid);
-  g_file_info_set_attribute_uint32 (ret_file_info, "unix::mode", mode);
+  ret_file_info = _ostree_header_gfile_info_new (mode, uid, gid);
 
   if (S_ISREG (mode))
     {
@@ -1416,26 +1683,25 @@ zlib_file_header_parse (GVariant         *metadata,
   guint64 size;
   guint32 uid, gid, mode, rdev;
   const char *symlink_target;
-  gs_unref_object GFileInfo *ret_file_info = NULL;
-  gs_unref_variant GVariant *ret_xattrs = NULL;
+  g_autoptr(GFileInfo) ret_file_info = NULL;
+  g_autoptr(GVariant) ret_xattrs = NULL;
 
   g_variant_get (metadata, "(tuuuu&s@a(ayay))", &size,
                  &uid, &gid, &mode, &rdev,
                  &symlink_target, &ret_xattrs);
+  if (rdev != 0)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                   "Corrupted archive file; invalid rdev %u", GUINT32_FROM_BE (rdev));
+      goto out;
+    }
 
-  size = GUINT64_FROM_BE (size);
   uid = GUINT32_FROM_BE (uid);
   gid = GUINT32_FROM_BE (gid);
   mode = GUINT32_FROM_BE (mode);
-  rdev = GUINT32_FROM_BE (rdev);
+  ret_file_info = _ostree_header_gfile_info_new (mode, uid, gid);
 
-  ret_file_info = g_file_info_new ();
-  g_file_info_set_size (ret_file_info, size);
-  g_file_info_set_attribute_uint32 (ret_file_info, "standard::type", ot_gfile_type_for_mode (mode));
-  g_file_info_set_attribute_boolean (ret_file_info, "standard::is-symlink", S_ISLNK (mode));
-  g_file_info_set_attribute_uint32 (ret_file_info, "unix::uid", uid);
-  g_file_info_set_attribute_uint32 (ret_file_info, "unix::gid", gid);
-  g_file_info_set_attribute_uint32 (ret_file_info, "unix::mode", mode);
+  g_file_info_set_size (ret_file_info, GUINT64_FROM_BE (size));
 
   if (S_ISREG (mode))
     {
@@ -1509,7 +1775,7 @@ ostree_validate_structureof_checksum_string (const char *checksum,
   int i = 0;
   size_t len = strlen (checksum);
 
-  if (len != 64)
+  if (len != OSTREE_SHA256_STRING_LEN)
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
                    "Invalid rev '%s'", checksum);
@@ -1568,9 +1834,9 @@ ostree_validate_structureof_commit (GVariant      *commit,
                                     GError       **error)
 {
   gboolean ret = FALSE;
-  gs_unref_variant GVariant *parent_csum_v = NULL;
-  gs_unref_variant GVariant *content_csum_v = NULL;
-  gs_unref_variant GVariant *metadata_csum_v = NULL;
+  g_autoptr(GVariant) parent_csum_v = NULL;
+  g_autoptr(GVariant) content_csum_v = NULL;
+  g_autoptr(GVariant) metadata_csum_v = NULL;
   gsize n_elts;
 
   if (!validate_variant (commit, OSTREE_COMMIT_GVARIANT_FORMAT, error))
@@ -1613,8 +1879,8 @@ ostree_validate_structureof_dirtree (GVariant      *dirtree,
 {
   gboolean ret = FALSE;
   const char *filename;
-  gs_unref_variant GVariant *content_csum_v = NULL;
-  gs_unref_variant GVariant *meta_csum_v = NULL;
+  g_autoptr(GVariant) content_csum_v = NULL;
+  g_autoptr(GVariant) meta_csum_v = NULL;
   GVariantIter *contents_iter = NULL;
 
   if (!validate_variant (dirtree, OSTREE_TREE_GVARIANT_FORMAT, error))
@@ -1743,13 +2009,14 @@ ostree_validate_structureof_dirmeta (GVariant      *dirmeta,
 /**
  * ostree_commit_get_parent:
  * @commit_variant: Variant of type %OSTREE_OBJECT_TYPE_COMMIT
- * 
- * Returns: Binary checksum with parent of @commit_variant, or %NULL if none
+ *
+ * Returns: Checksum of the parent commit of @commit_variant, or %NULL
+ * if none
  */
 gchar *
 ostree_commit_get_parent (GVariant  *commit_variant)
 {
-  gs_unref_variant GVariant *bytes = NULL;
+  g_autoptr(GVariant) bytes = NULL;
   bytes = g_variant_get_child_value (commit_variant, 1);
   if (g_variant_n_children (bytes) == 0)
     return NULL;
@@ -1763,3 +2030,56 @@ ostree_commit_get_timestamp (GVariant  *commit_variant)
   g_variant_get_child (commit_variant, 5, "t", &ret);
   return GUINT64_FROM_BE (ret);
 }
+
+GVariant *
+_ostree_detached_metadata_append_gpg_sig (GVariant   *existing_metadata,
+                                          GBytes     *signature_bytes)
+{
+  GVariantDict metadata_dict;
+  g_autoptr(GVariant) signature_data = NULL;
+  g_autoptr(GVariantBuilder) signature_builder = NULL;
+
+  g_variant_dict_init (&metadata_dict, existing_metadata);
+
+  signature_data = g_variant_dict_lookup_value (&metadata_dict,
+                                                _OSTREE_METADATA_GPGSIGS_NAME,
+                                                _OSTREE_METADATA_GPGSIGS_TYPE);
+
+  /* signature_data may be NULL */
+  signature_builder = ot_util_variant_builder_from_variant (signature_data, _OSTREE_METADATA_GPGSIGS_TYPE);
+
+  g_variant_builder_add (signature_builder, "@ay", ot_gvariant_new_ay_bytes (signature_bytes));
+
+  g_variant_dict_insert_value (&metadata_dict,
+                               _OSTREE_METADATA_GPGSIGS_NAME,
+                               g_variant_builder_end (signature_builder));
+
+  return g_variant_dict_end (&metadata_dict);
+}
+
+/**
+ * _ostree_get_default_sysroot_path:
+ *
+ * Returns a #GFile for the default system root, which is usually the root
+ * directory ("/") unless overridden by the %OSTREE_SYSROOT environment
+ * variable.
+ *
+ * Returns: a #GFile for the default system root
+ */
+GFile *
+_ostree_get_default_sysroot_path (void)
+{
+  static gsize default_sysroot_path_initialized;
+  static GFile *default_sysroot_path;
+
+  if (g_once_init_enter (&default_sysroot_path_initialized))
+    {
+      const char *path = g_getenv ("OSTREE_SYSROOT");
+      if (path == NULL || *path == '\0')
+        path = "/";
+      default_sysroot_path = g_file_new_for_path (path);
+      g_once_init_leave (&default_sysroot_path_initialized, 1);
+    }
+
+  return default_sysroot_path;
+}
