@@ -20,22 +20,8 @@
 
 #include "config.h"
 
-#include "ostree-deployment.h"
-#include "libgsystem.h"
-
-struct _OstreeDeployment
-{
-  GObject       parent_instance;
-
-  int index;  /* Global offset */
-  char *osname;  /* osname */
-  char *csum;  /* OSTree checksum of tree */
-  int deployserial;  /* How many times this particular csum appears in deployment list */
-  char *bootcsum;  /* Checksum of kernel+initramfs */
-  int bootserial; /* An integer assigned to this tree per its ${bootcsum} */
-  OstreeBootconfigParser *bootconfig; /* Bootloader configuration */
-  GKeyFile *origin; /* How to construct an upgraded version of this tree */
-};
+#include "ostree-deployment-private.h"
+#include "libglnx.h"
 
 typedef GObjectClass OstreeDeploymentClass;
 
@@ -129,6 +115,14 @@ ostree_deployment_set_origin (OstreeDeployment *self, GKeyFile *origin)
     self->origin = g_key_file_ref (origin);
 }
 
+void
+_ostree_deployment_set_bootcsum (OstreeDeployment *self,
+                                 const char *bootcsum)
+{
+  g_free (self->bootcsum);
+  self->bootcsum = g_strdup (bootcsum);
+}
+
 /**
  * ostree_deployment_clone:
  * @self: Deployment
@@ -138,7 +132,7 @@ ostree_deployment_set_origin (OstreeDeployment *self, GKeyFile *origin)
 OstreeDeployment *
 ostree_deployment_clone (OstreeDeployment *self)
 {
-  gs_unref_object OstreeBootconfigParser *new_bootconfig = NULL;
+  glnx_unref_object OstreeBootconfigParser *new_bootconfig = NULL;
   GKeyFile *new_origin = NULL;
   OstreeDeployment *ret = ostree_deployment_new (self->index, self->osname, self->csum,
                                                  self->deployserial,
@@ -149,7 +143,7 @@ ostree_deployment_clone (OstreeDeployment *self)
 
   if (self->origin)
     {
-      gs_free char *data = NULL;
+      g_autofree char *data = NULL;
       gsize len;
       gboolean success;
 
@@ -250,5 +244,46 @@ ostree_deployment_new (int    index,
   self->deployserial = deployserial;
   self->bootcsum = g_strdup (bootcsum);
   self->bootserial = bootserial;
+  self->unlocked = OSTREE_DEPLOYMENT_UNLOCKED_NONE;
   return self;
 }
+
+/**
+ * ostree_deployment_get_origin_relpath:
+ * @self: A deployment
+ *
+ * Note this function only returns a *relative* path - if you want to
+ * access, it, you must either use fd-relative api such as openat(),
+ * or concatenate it with the full ostree_sysroot_get_path().
+ *
+ * Returns: (transfer full): Path to deployment root directory, relative to sysroot
+ */
+char *
+ostree_deployment_get_origin_relpath (OstreeDeployment *self)
+{
+  return g_strdup_printf ("ostree/deploy/%s/deploy/%s.%d.origin",
+                          ostree_deployment_get_osname (self),
+                          ostree_deployment_get_csum (self),
+                          ostree_deployment_get_deployserial (self));
+}
+
+const char *
+ostree_deployment_unlocked_state_to_string (OstreeDeploymentUnlockedState state)
+{
+  switch (state)
+    {
+    case OSTREE_DEPLOYMENT_UNLOCKED_NONE:
+      return "none";
+    case OSTREE_DEPLOYMENT_UNLOCKED_HOTFIX:
+      return "hotfix";
+    case OSTREE_DEPLOYMENT_UNLOCKED_DEVELOPMENT:
+      return "development";
+    }
+  g_assert_not_reached ();
+}
+
+OstreeDeploymentUnlockedState
+ostree_deployment_get_unlocked (OstreeDeployment *self)
+{
+  return self->unlocked;
+}
