@@ -23,7 +23,7 @@
 #include "config.h"
 
 #include <glib-unix.h>
-#include <attr/xattr.h>
+#include <sys/xattr.h>
 #include <gio/gfiledescriptorbased.h>
 #include <gio/gunixoutputstream.h>
 #include "otutil.h"
@@ -32,6 +32,8 @@
 #include "ostree-core-private.h"
 #include "ostree-repo-private.h"
 
+#define WHITEOUT_PREFIX ".wh."
+
 static gboolean
 checkout_object_for_uncompressed_cache (OstreeRepo      *self,
                                         const char      *loose_path,
@@ -41,9 +43,9 @@ checkout_object_for_uncompressed_cache (OstreeRepo      *self,
                                         GError         **error)
 {
   gboolean ret = FALSE;
-  gs_free char *temp_filename = NULL;
-  gs_unref_object GOutputStream *temp_out = NULL;
-  int fd;
+  g_autofree char *temp_filename = NULL;
+  g_autoptr(GOutputStream) temp_out = NULL;
+  glnx_fd_close int fd = -1;
   int res;
   guint32 file_mode;
 
@@ -51,10 +53,11 @@ checkout_object_for_uncompressed_cache (OstreeRepo      *self,
   file_mode = g_file_info_get_attribute_uint32 (src_info, "unix::mode");
   file_mode &= ~(S_ISUID|S_ISGID);
 
-  if (!gs_file_open_in_tmpdir_at (self->tmp_dir_fd, file_mode,
-                                  &temp_filename, &temp_out,
-                                  cancellable, error))
+  if (!glnx_open_tmpfile_linkable_at (self->tmp_dir_fd, ".", O_WRONLY | O_CLOEXEC,
+                                      &fd, &temp_filename,
+                                      error))
     goto out;
+  temp_out = g_unix_output_stream_new (fd, FALSE);
 
   if (g_output_stream_splice (temp_out, content, 0, cancellable, error) < 0)
     goto out;
@@ -62,37 +65,37 @@ checkout_object_for_uncompressed_cache (OstreeRepo      *self,
   if (!g_output_stream_flush (temp_out, cancellable, error))
     goto out;
 
-  fd = g_file_descriptor_based_get_fd ((GFileDescriptorBased*)temp_out);
-
-  do
-    res = fsync (fd);
-  while (G_UNLIKELY (res == -1 && errno == EINTR));
-  if (G_UNLIKELY (res == -1))
+  if (!self->disable_fsync)
     {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
+      do
+        res = fsync (fd);
+      while (G_UNLIKELY (res == -1 && errno == EINTR));
+      if (G_UNLIKELY (res == -1))
+        {
+          glnx_set_error_from_errno (error);
+          goto out;
+        }
     }
 
   if (!g_output_stream_close (temp_out, cancellable, error))
     goto out;
 
+  if (fchmod (fd, file_mode) < 0)
+    {
+      glnx_set_error_from_errno (error);
+      goto out;
+    }
+
   if (!_ostree_repo_ensure_loose_objdir_at (self->uncompressed_objects_dir_fd,
                                             loose_path,
                                             cancellable, error))
     goto out;
 
-  if (G_UNLIKELY (renameat (self->tmp_dir_fd, temp_filename,
-                            self->uncompressed_objects_dir_fd, loose_path) == -1))
-    {
-      if (errno != EEXIST)
-        {
-          ot_util_set_error_from_errno (error, errno);
-          g_prefix_error (error, "Storing file '%s': ", temp_filename);
-          goto out;
-        }
-      else
-        (void) unlinkat (self->tmp_dir_fd, temp_filename, 0);
-    }
+  if (!glnx_link_tmpfile_at (self->tmp_dir_fd, GLNX_LINK_TMPFILE_NOREPLACE_IGNORE_EXIST,
+                             fd, temp_filename,
+                             self->uncompressed_objects_dir_fd, loose_path,
+                             error))
+    goto out;
 
   ret = TRUE;
  out:
@@ -100,7 +103,15 @@ checkout_object_for_uncompressed_cache (OstreeRepo      *self,
 }
 
 static gboolean
-write_regular_file_content (OstreeRepoCheckoutMode mode,
+fsync_is_enabled (OstreeRepo   *self,
+                  OstreeRepoCheckoutAtOptions *options)
+{
+  return options->enable_fsync || !self->disable_fsync;
+}
+
+static gboolean
+write_regular_file_content (OstreeRepo            *self,
+                            OstreeRepoCheckoutAtOptions *options,
                             GOutputStream         *output,
                             GFileInfo             *file_info,
                             GVariant              *xattrs,
@@ -109,6 +120,7 @@ write_regular_file_content (OstreeRepoCheckoutMode mode,
                             GError               **error)
 {
   gboolean ret = FALSE;
+  const OstreeRepoCheckoutMode mode = options->mode;
   int fd;
   int res;
 
@@ -130,7 +142,7 @@ write_regular_file_content (OstreeRepoCheckoutMode mode,
       while (G_UNLIKELY (res == -1 && errno == EINTR));
       if (G_UNLIKELY (res == -1))
         {
-          ot_util_set_error_from_errno (error, errno);
+          glnx_set_error_from_errno (error);
           goto out;
         }
 
@@ -139,21 +151,24 @@ write_regular_file_content (OstreeRepoCheckoutMode mode,
       while (G_UNLIKELY (res == -1 && errno == EINTR));
       if (G_UNLIKELY (res == -1))
         {
-          ot_util_set_error_from_errno (error, errno);
+          glnx_set_error_from_errno (error);
           goto out;
         }
               
       if (xattrs)
         {
-          if (!gs_fd_set_all_xattrs (fd, xattrs, cancellable, error))
+          if (!glnx_fd_set_all_xattrs (fd, xattrs, cancellable, error))
             goto out;
         }
     }
           
-  if (fsync (fd) == -1)
+  if (fsync_is_enabled (self, options))
     {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
+      if (fsync (fd) == -1)
+        {
+          glnx_set_error_from_errno (error);
+          goto out;
+        }
     }
           
   if (!g_output_stream_close (output, cancellable, error))
@@ -165,12 +180,12 @@ write_regular_file_content (OstreeRepoCheckoutMode mode,
 }
 
 static gboolean
-checkout_file_from_input_at (OstreeRepoCheckoutMode mode,
+checkout_file_from_input_at (OstreeRepo     *self,
+                             OstreeRepoCheckoutAtOptions *options,
                              GFileInfo      *file_info,
                              GVariant       *xattrs,
                              GInputStream   *input,
                              int             destination_dfd,
-                             GFile          *destination_parent,
                              const char     *destination_name,
                              GCancellable   *cancellable,
                              GError        **error)
@@ -186,26 +201,38 @@ checkout_file_from_input_at (OstreeRepoCheckoutMode mode,
       while (G_UNLIKELY (res == -1 && errno == EINTR));
       if (res == -1)
         {
-          ot_util_set_error_from_errno (error, errno);
+          glnx_set_error_from_errno (error);
           goto out;
         }
-          
-      if (xattrs)
+
+      if (options->mode != OSTREE_REPO_CHECKOUT_MODE_USER)
         {
-          if (!gs_dfd_and_name_set_all_xattrs (destination_dfd, destination_name,
-                                               xattrs, cancellable, error))
-            goto out;
+          if (G_UNLIKELY (fchownat (destination_dfd, destination_name,
+                                    g_file_info_get_attribute_uint32 (file_info, "unix::uid"),
+                                    g_file_info_get_attribute_uint32 (file_info, "unix::gid"),
+                                    AT_SYMLINK_NOFOLLOW) == -1))
+            {
+              glnx_set_error_from_errno (error);
+              goto out;
+            }
+
+          if (xattrs)
+            {
+              if (!glnx_dfd_name_set_all_xattrs (destination_dfd, destination_name,
+                                                   xattrs, cancellable, error))
+                goto out;
+            }
         }
     }
   else if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_REGULAR)
     {
-      gs_unref_object GOutputStream *temp_out = NULL;
+      g_autoptr(GOutputStream) temp_out = NULL;
       int fd;
       guint32 file_mode;
 
       file_mode = g_file_info_get_attribute_uint32 (file_info, "unix::mode");
       /* Don't make setuid files on checkout when we're doing --user */
-      if (mode == OSTREE_REPO_CHECKOUT_MODE_USER)
+      if (options->mode == OSTREE_REPO_CHECKOUT_MODE_USER)
         file_mode &= ~(S_ISUID|S_ISGID);
 
       do
@@ -213,13 +240,13 @@ checkout_file_from_input_at (OstreeRepoCheckoutMode mode,
       while (G_UNLIKELY (fd == -1 && errno == EINTR));
       if (fd == -1)
         {
-          ot_util_set_error_from_errno (error, errno);
+          glnx_set_error_from_errno (error);
           goto out;
         }
       temp_out = g_unix_output_stream_new (fd, TRUE);
       fd = -1; /* Transfer ownership */
 
-      if (!write_regular_file_content (mode, temp_out, file_info, xattrs, input,
+      if (!write_regular_file_content (self, options, temp_out, file_info, xattrs, input,
                                        cancellable, error))
         goto out;
     }
@@ -236,18 +263,18 @@ checkout_file_from_input_at (OstreeRepoCheckoutMode mode,
  * it into place.  This implements union-like behavior.
  */
 static gboolean
-checkout_file_unioning_from_input_at (OstreeRepoCheckoutMode mode,
+checkout_file_unioning_from_input_at (OstreeRepo     *repo,
+                                      OstreeRepoCheckoutAtOptions  *options,
                                       GFileInfo      *file_info,
                                       GVariant       *xattrs,
                                       GInputStream   *input,
                                       int             destination_dfd,
-                                      GFile          *destination_parent,
                                       const char     *destination_name,
                                       GCancellable   *cancellable,
                                       GError        **error)
 {
   gboolean ret = FALSE;
-  gs_free char *temp_filename = NULL;
+  g_autofree char *temp_filename = NULL;
 
   if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_SYMBOLIC_LINK)
     {
@@ -259,40 +286,47 @@ checkout_file_unioning_from_input_at (OstreeRepoCheckoutMode mode,
           
       if (xattrs)
         {
-          if (!gs_dfd_and_name_set_all_xattrs (destination_dfd, destination_name,
+          if (!glnx_dfd_name_set_all_xattrs (destination_dfd, temp_filename,
                                                xattrs, cancellable, error))
             goto out;
         }
+      if (G_UNLIKELY (renameat (destination_dfd, temp_filename,
+                                destination_dfd, destination_name) == -1))
+        {
+          glnx_set_error_from_errno (error);
+          goto out;
+        }
     }
   else if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_REGULAR)
     {
-      gs_unref_object GOutputStream *temp_out = NULL;
+      glnx_fd_close int temp_fd = -1;
+      g_autoptr(GOutputStream) temp_out = NULL;
       guint32 file_mode;
 
       file_mode = g_file_info_get_attribute_uint32 (file_info, "unix::mode");
       /* Don't make setuid files on checkout when we're doing --user */
-      if (mode == OSTREE_REPO_CHECKOUT_MODE_USER)
+      if (options->mode == OSTREE_REPO_CHECKOUT_MODE_USER)
         file_mode &= ~(S_ISUID|S_ISGID);
 
-      if (!gs_file_open_in_tmpdir_at (destination_dfd, file_mode,
-                                      &temp_filename, &temp_out,
-                                      cancellable, error))
+      if (!glnx_open_tmpfile_linkable_at (destination_dfd, ".", O_WRONLY | O_CLOEXEC,
+                                          &temp_fd, &temp_filename,
+                                          error))
+        goto out;
+      temp_out = g_unix_output_stream_new (temp_fd, FALSE);
+
+      if (!write_regular_file_content (repo, options, temp_out, file_info, xattrs, input,
+                                       cancellable, error))
         goto out;
 
-      if (!write_regular_file_content (mode, temp_out, file_info, xattrs, input,
-                                       cancellable, error))
+      if (!glnx_link_tmpfile_at (destination_dfd, GLNX_LINK_TMPFILE_REPLACE,
+                                 temp_fd, temp_filename, destination_dfd,
+                                 destination_name,
+                                 error))
         goto out;
     }
   else
     g_assert_not_reached ();
 
-  if (G_UNLIKELY (renameat (destination_dfd, temp_filename,
-                            destination_dfd, destination_name) == -1))
-    {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
-    }
-
   ret = TRUE;
  out:
   return ret;
@@ -300,8 +334,7 @@ checkout_file_unioning_from_input_at (OstreeRepoCheckoutMode mode,
 
 static gboolean
 checkout_file_hardlink (OstreeRepo                          *self,
-                        OstreeRepoCheckoutMode               mode,
-                        OstreeRepoCheckoutOverwriteMode      overwrite_mode,
+                        OstreeRepoCheckoutAtOptions           *options,
                         const char                          *loose_path,
                         int                                  destination_dfd,
                         const char                          *destination_name,
@@ -312,13 +345,13 @@ checkout_file_hardlink (OstreeRepo                          *self,
 {
   gboolean ret = FALSE;
   gboolean ret_was_supported = FALSE;
-  int srcfd = self->mode == OSTREE_REPO_MODE_BARE ?
+  int srcfd = (self->mode == OSTREE_REPO_MODE_BARE || self->mode == OSTREE_REPO_MODE_BARE_USER) ?
     self->objects_dir_fd : self->uncompressed_objects_dir_fd;
 
  again:
   if (linkat (srcfd, loose_path, destination_dfd, destination_name, 0) != -1)
     ret_was_supported = TRUE;
-  else if (errno == EMLINK || errno == EXDEV || errno == EPERM)
+  else if (!options->no_copy_fallback && (errno == EMLINK || errno == EXDEV || errno == EPERM))
     {
       /* EMLINK, EXDEV and EPERM shouldn't be fatal; we just can't do the
        * optimization of hardlinking instead of copying.
@@ -329,7 +362,7 @@ checkout_file_hardlink (OstreeRepo                          *self,
     {
       ret_was_supported = FALSE;
     }
-  else if (errno == EEXIST && overwrite_mode == OSTREE_REPO_CHECKOUT_OVERWRITE_UNION_FILES)
+  else if (errno == EEXIST && options->overwrite_mode == OSTREE_REPO_CHECKOUT_OVERWRITE_UNION_FILES)
     { 
       /* Idiocy, from man rename(2)
        *
@@ -345,7 +378,7 @@ checkout_file_hardlink (OstreeRepo                          *self,
   else
     {
       g_prefix_error (error, "Hardlinking %s to %s: ", loose_path, destination_name);
-      ot_util_set_error_from_errno (error, errno);
+      glnx_set_error_from_errno (error);
       goto out;
     }
 
@@ -358,41 +391,71 @@ checkout_file_hardlink (OstreeRepo                          *self,
 
 static gboolean
 checkout_one_file_at (OstreeRepo                        *repo,
+                      OstreeRepoCheckoutAtOptions         *options,
                       GFile                             *source,
                       GFileInfo                         *source_info,
                       int                                destination_dfd,
-                      GFile                             *destination_parent,
                       const char                        *destination_name,
-                      OstreeRepoCheckoutMode             mode,
-                      OstreeRepoCheckoutOverwriteMode    overwrite_mode,
                       GCancellable                      *cancellable,
                       GError                           **error)
 {
   gboolean ret = FALSE;
   const char *checksum;
   gboolean is_symlink;
-  gboolean did_hardlink = FALSE;
+  gboolean can_cache;
+  gboolean need_copy = TRUE;
   char loose_path_buf[_OSTREE_LOOSE_PATH_MAX];
-  gs_unref_object GInputStream *input = NULL;
-  gs_unref_variant GVariant *xattrs = NULL;
+  g_autoptr(GInputStream) input = NULL;
+  g_autoptr(GVariant) xattrs = NULL;
+  gboolean is_whiteout;
 
   is_symlink = g_file_info_get_file_type (source_info) == G_FILE_TYPE_SYMBOLIC_LINK;
 
   checksum = ostree_repo_file_get_checksum ((OstreeRepoFile*)source);
 
-  /* Try to do a hardlink first, if it's a regular file.  This also
-   * traverses all parent repos.
+  is_whiteout = !is_symlink && options->process_whiteouts &&
+    g_str_has_prefix (destination_name, WHITEOUT_PREFIX);
+
+  /* First, see if it's a Docker whiteout,
+   * https://github.com/docker/docker/blob/1a714e76a2cb9008cd19609059e9988ff1660b78/pkg/archive/whiteouts.go
    */
-  if (!is_symlink)
+  if (is_whiteout)
     {
+      const char *name = destination_name + (sizeof (WHITEOUT_PREFIX) - 1);
+
+      if (!name[0])
+        {
+          g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                       "Invalid empty whiteout '%s'", name);
+          goto out;
+        }
+
+      g_assert (name[0] != '/'); /* Sanity */
+
+      if (!glnx_shutil_rm_rf_at (destination_dfd, name, cancellable, error))
+        goto out;
+
+      need_copy = FALSE;
+    }
+  else if (!is_symlink)
+    {
+      gboolean did_hardlink = FALSE;
+      /* Try to do a hardlink first, if it's a regular file.  This also
+       * traverses all parent repos.
+       */
       OstreeRepo *current_repo = repo;
 
       while (current_repo)
         {
-          gboolean is_bare = (current_repo->mode == OSTREE_REPO_MODE_BARE
-                              && mode == OSTREE_REPO_CHECKOUT_MODE_NONE);
+          gboolean is_bare = ((current_repo->mode == OSTREE_REPO_MODE_BARE
+                               && options->mode == OSTREE_REPO_CHECKOUT_MODE_NONE) ||
+                              (current_repo->mode == OSTREE_REPO_MODE_BARE_USER
+                               && options->mode == OSTREE_REPO_CHECKOUT_MODE_USER));
+          gboolean current_can_cache = (options->enable_uncompressed_cache
+                                        && current_repo->enable_uncompressed_cache);
           gboolean is_archive_z2_with_cache = (current_repo->mode == OSTREE_REPO_MODE_ARCHIVE_Z2
-                                               && mode == OSTREE_REPO_CHECKOUT_MODE_USER);
+                                               && options->mode == OSTREE_REPO_CHECKOUT_MODE_USER
+                                               && current_can_cache);
 
           /* But only under these conditions */
           if (is_bare || is_archive_z2_with_cache)
@@ -401,27 +464,56 @@ checkout_one_file_at (OstreeRepo                        *repo,
                  the cache, which is in "bare" form */
               _ostree_loose_path (loose_path_buf, checksum, OSTREE_OBJECT_TYPE_FILE, OSTREE_REPO_MODE_BARE);
               if (!checkout_file_hardlink (current_repo,
-                                           mode, overwrite_mode, loose_path_buf,
+                                           options,
+                                           loose_path_buf,
                                            destination_dfd, destination_name,
                                            TRUE, &did_hardlink,
                                            cancellable, error))
                 goto out;
+
+              if (did_hardlink && options->devino_to_csum_cache)
+                {
+                  struct stat stbuf;
+                  OstreeDevIno *key;
+                  
+                  if (TEMP_FAILURE_RETRY (fstatat (destination_dfd, destination_name, &stbuf, AT_SYMLINK_NOFOLLOW)) != 0)
+                    {
+                      glnx_set_error_from_errno (error);
+                      goto out;
+                    }
+                  
+                  key = g_new (OstreeDevIno, 1);
+                  key->dev = stbuf.st_dev;
+                  key->ino = stbuf.st_ino;
+                  memcpy (key->checksum, checksum, OSTREE_SHA256_STRING_LEN+1);
+                  
+                  g_hash_table_add ((GHashTable*)options->devino_to_csum_cache, key);
+                }
+
               if (did_hardlink)
                 break;
             }
           current_repo = current_repo->parent_repo;
         }
+
+      need_copy = !did_hardlink;
     }
 
+  can_cache = (options->enable_uncompressed_cache
+               && repo->enable_uncompressed_cache);
+
   /* Ok, if we're archive-z2 and we didn't find an object, uncompress
    * it now, stick it in the cache, and then hardlink to that.
    */
-  if (!is_symlink
-      && !did_hardlink
+  if (can_cache
+      && !is_whiteout
+      && !is_symlink
+      && need_copy
       && repo->mode == OSTREE_REPO_MODE_ARCHIVE_Z2
-      && mode == OSTREE_REPO_CHECKOUT_MODE_USER
-      && repo->enable_uncompressed_cache)
+      && options->mode == OSTREE_REPO_CHECKOUT_MODE_USER)
     {
+      gboolean did_hardlink;
+      
       if (!ostree_repo_load_file (repo, checksum, &input, NULL, NULL,
                                   cancellable, error))
         goto out;
@@ -464,7 +556,7 @@ checkout_one_file_at (OstreeRepo                        *repo,
       }
       g_mutex_unlock (&repo->cache_lock);
 
-      if (!checkout_file_hardlink (repo, mode, overwrite_mode, loose_path_buf,
+      if (!checkout_file_hardlink (repo, options, loose_path_buf,
                                    destination_dfd, destination_name,
                                    FALSE, &did_hardlink,
                                    cancellable, error))
@@ -472,19 +564,21 @@ checkout_one_file_at (OstreeRepo                        *repo,
           g_prefix_error (error, "Using new cached uncompressed hardlink of %s to %s: ", checksum, destination_name);
           goto out;
         }
+
+      need_copy = !did_hardlink;
     }
 
   /* Fall back to copy if we couldn't hardlink */
-  if (!did_hardlink)
+  if (need_copy)
     {
       if (!ostree_repo_load_file (repo, checksum, &input, NULL, &xattrs,
                                   cancellable, error))
         goto out;
 
-      if (overwrite_mode == OSTREE_REPO_CHECKOUT_OVERWRITE_UNION_FILES)
+      if (options->overwrite_mode == OSTREE_REPO_CHECKOUT_OVERWRITE_UNION_FILES)
         {
-          if (!checkout_file_unioning_from_input_at (mode, source_info, xattrs, input,
-                                                     destination_dfd, destination_parent,
+          if (!checkout_file_unioning_from_input_at (repo, options, source_info, xattrs, input,
+                                                     destination_dfd,
                                                      destination_name,
                                                      cancellable, error)) 
             {
@@ -494,8 +588,8 @@ checkout_one_file_at (OstreeRepo                        *repo,
         }
       else
         {
-          if (!checkout_file_from_input_at (mode, source_info, xattrs, input,
-                                            destination_dfd, destination_parent,
+          if (!checkout_file_from_input_at (repo, options, source_info, xattrs, input,
+                                            destination_dfd,
                                             destination_name,
                                             cancellable, error))
             {
@@ -533,11 +627,9 @@ checkout_one_file_at (OstreeRepo                        *repo,
  */
 static gboolean
 checkout_tree_at (OstreeRepo                        *self,
-                  OstreeRepoCheckoutMode             mode,
-                  OstreeRepoCheckoutOverwriteMode    overwrite_mode,
+                  OstreeRepoCheckoutAtOptions       *options,
                   int                                destination_parent_fd,
                   const char                        *destination_name,
-                  GFile                             *destination,
                   OstreeRepoFile                    *source,
                   GFileInfo                         *source_info,
                   GCancellable                      *cancellable,
@@ -545,10 +637,10 @@ checkout_tree_at (OstreeRepo                        *self,
 {
   gboolean ret = FALSE;
   gboolean did_exist = FALSE;
-  int destination_dfd = -1;
+  glnx_fd_close int destination_dfd = -1;
   int res;
-  gs_unref_variant GVariant *xattrs = NULL;
-  gs_unref_object GFileEnumerator *dir_enum = NULL;
+  g_autoptr(GVariant) xattrs = NULL;
+  g_autoptr(GFileEnumerator) dir_enum = NULL;
 
   /* Create initially with mode 0700, then chown/chmod only when we're
    * done.  This avoids anyone else being able to operate on partially
@@ -559,33 +651,42 @@ checkout_tree_at (OstreeRepo                        *self,
   while (G_UNLIKELY (res == -1 && errno == EINTR));
   if (res == -1)
     {
-      if (errno == EEXIST && overwrite_mode == OSTREE_REPO_CHECKOUT_OVERWRITE_UNION_FILES)
+      if (errno == EEXIST && options->overwrite_mode == OSTREE_REPO_CHECKOUT_OVERWRITE_UNION_FILES)
         did_exist = TRUE;
       else
         {
-          ot_util_set_error_from_errno (error, errno);
+          glnx_set_error_from_errno (error);
           goto out;
         }
     }
 
-  if (!gs_file_open_dir_fd_at (destination_parent_fd, destination_name,
-                               &destination_dfd,
-                               cancellable, error))
+  if (!glnx_opendirat (destination_parent_fd, destination_name, TRUE,
+                       &destination_dfd, error))
     goto out;
 
   /* Set the xattrs now, so any derived labeling works */
-  if (!did_exist && mode != OSTREE_REPO_CHECKOUT_MODE_USER)
+  if (!did_exist && options->mode != OSTREE_REPO_CHECKOUT_MODE_USER)
     {
       if (!ostree_repo_file_get_xattrs (source, &xattrs, NULL, error))
         goto out;
 
       if (xattrs)
         {
-          if (!gs_fd_set_all_xattrs (destination_dfd, xattrs, cancellable, error))
+          if (!glnx_fd_set_all_xattrs (destination_dfd, xattrs, cancellable, error))
             goto out;
         }
     }
 
+  if (g_file_info_get_file_type (source_info) != G_FILE_TYPE_DIRECTORY)
+    {
+      ret = checkout_one_file_at (self, options,
+                                  (GFile *) source,
+                                  source_info,
+                                  destination_dfd,
+                                  g_file_info_get_name (source_info),
+                                  cancellable, error);
+      goto out;
+    }
   dir_enum = g_file_enumerate_children ((GFile*)source,
                                         OSTREE_GIO_FAST_QUERYINFO, 
                                         G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
@@ -600,8 +701,8 @@ checkout_tree_at (OstreeRepo                        *self,
       GFile *src_child;
       const char *name;
 
-      if (!gs_file_enumerator_iterate (dir_enum, &file_info, &src_child,
-                                       cancellable, error))
+      if (!g_file_enumerator_iterate (dir_enum, &file_info, &src_child,
+                                      cancellable, error))
         goto out;
       if (file_info == NULL)
         break;
@@ -610,18 +711,17 @@ checkout_tree_at (OstreeRepo                        *self,
 
       if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_DIRECTORY)
         {
-          gs_unref_object GFile *child_destination = g_file_get_child (destination, name);
-          if (!checkout_tree_at (self, mode, overwrite_mode,
-                                 destination_dfd, name, child_destination,
+          if (!checkout_tree_at (self, options,
+                                 destination_dfd, name,
                                  (OstreeRepoFile*)src_child, file_info,
                                  cancellable, error))
             goto out;
         }
       else
         {
-          if (!checkout_one_file_at (self, src_child, file_info,
-                                     destination_dfd, destination, name,
-                                     mode, overwrite_mode,
+          if (!checkout_one_file_at (self, options,
+                                     src_child, file_info,
+                                     destination_dfd, name,
                                      cancellable, error))
             goto out;
         }
@@ -638,12 +738,12 @@ checkout_tree_at (OstreeRepo                        *self,
       while (G_UNLIKELY (res == -1 && errno == EINTR));
       if (G_UNLIKELY (res == -1))
         {
-          ot_util_set_error_from_errno (error, errno);
+          glnx_set_error_from_errno (error);
           goto out;
         }
     }
 
-  if (!did_exist && mode != OSTREE_REPO_CHECKOUT_MODE_USER)
+  if (!did_exist && options->mode != OSTREE_REPO_CHECKOUT_MODE_USER)
     {
       do
         res = fchown (destination_dfd,
@@ -652,25 +752,38 @@ checkout_tree_at (OstreeRepo                        *self,
       while (G_UNLIKELY (res == -1 && errno == EINTR));
       if (G_UNLIKELY (res == -1))
         {
-          ot_util_set_error_from_errno (error, errno);
+          glnx_set_error_from_errno (error);
           goto out;
         }
     }
 
-  /* Finally, fsync to ensure all entries are on disk.  Ultimately
-   * this should be configurable for the case where we're constructing
-   * buildroots.
+  /* Set directory mtime to OSTREE_TIMESTAMP, so that it is constant for all checkouts.
+   * Must be done after setting permissions and creating all children.
    */
-  if (fsync (destination_dfd) == -1)
+  if (!did_exist)
     {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
+      const struct timespec times[2] = { { OSTREE_TIMESTAMP, UTIME_OMIT }, { OSTREE_TIMESTAMP, 0} };
+      do
+        res = futimens (destination_dfd, times);
+      while (G_UNLIKELY (res == -1 && errno == EINTR));
+      if (G_UNLIKELY (res == -1))
+        {
+          glnx_set_error_from_errno (error);
+          goto out;
+        }
+    }
+
+  if (fsync_is_enabled (self, options))
+    {
+      if (fsync (destination_dfd) == -1)
+        {
+          glnx_set_error_from_errno (error);
+          goto out;
+        }
     }
 
   ret = TRUE;
  out:
-  if (destination_dfd != -1)
-    (void) close (destination_dfd);
   return ret;
 }
 
@@ -700,14 +813,170 @@ ostree_repo_checkout_tree (OstreeRepo               *self,
                            GCancellable             *cancellable,
                            GError                  **error)
 {
-  return checkout_tree_at (self, mode, overwrite_mode,
-                           AT_FDCWD,
-                           gs_file_get_path_cached (destination),
-                           destination,
+  OstreeRepoCheckoutAtOptions options = { 0, };
+
+  options.mode = mode;
+  options.overwrite_mode = overwrite_mode;
+  /* Backwards compatibility */
+  options.enable_uncompressed_cache = TRUE;
+
+  return checkout_tree_at (self, &options,
+                           AT_FDCWD, gs_file_get_path_cached (destination),
                            source, source_info,
                            cancellable, error);
 }
 
+/**
+ * ostree_repo_checkout_tree_at: (skip)
+ * @self: Repo
+ * @options: (allow-none): Options
+ * @destination_dfd: Directory FD for destination
+ * @destination_path: Directory for destination
+ * @commit: Checksum for commit
+ * @cancellable: Cancellable
+ * @error: Error
+ *
+ * Similar to ostree_repo_checkout_tree(), but uses directory-relative
+ * paths for the destination, uses a new `OstreeRepoCheckoutAtOptions`,
+ * and takes a commit checksum and optional subpath pair, rather than
+ * requiring use of `GFile` APIs for the caller.
+ *
+ * Note in addition that unlike ostree_repo_checkout_tree(), the
+ * default is not to use the repository-internal uncompressed objects
+ * cache.
+ *
+ * This function is deprecated.  Use ostree_repo_checkout_at() instead.
+ */
+gboolean
+ostree_repo_checkout_tree_at (OstreeRepo                        *self,
+                              OstreeRepoCheckoutOptions         *options,
+                              int                                destination_dfd,
+                              const char                        *destination_path,
+                              const char                        *commit,
+                              GCancellable                      *cancellable,
+                              GError                           **error)
+{
+  OstreeRepoCheckoutAtOptions new_opts = {0, };
+  new_opts.mode = options->mode;
+  new_opts.overwrite_mode = options->overwrite_mode;
+  new_opts.enable_uncompressed_cache = options->enable_uncompressed_cache;
+  new_opts.enable_fsync = !options->disable_fsync;
+  new_opts.process_whiteouts = options->process_whiteouts;
+  new_opts.no_copy_fallback = options->no_copy_fallback;
+  new_opts.subpath = options->subpath;
+  new_opts.devino_to_csum_cache = options->devino_to_csum_cache;
+  return ostree_repo_checkout_at (self, &new_opts, destination_dfd,
+                                  destination_path, commit, cancellable, error);
+}
+
+/**
+ * ostree_repo_checkout_at:
+ * @self: Repo
+ * @options: (allow-none): Options
+ * @destination_dfd: Directory FD for destination
+ * @destination_path: Directory for destination
+ * @commit: Checksum for commit
+ * @cancellable: Cancellable
+ * @error: Error
+ *
+ * Similar to ostree_repo_checkout_tree(), but uses directory-relative
+ * paths for the destination, uses a new `OstreeRepoCheckoutAtOptions`,
+ * and takes a commit checksum and optional subpath pair, rather than
+ * requiring use of `GFile` APIs for the caller.
+ *
+ * It also replaces ostree_repo_checkout_at() which was not safe to
+ * use with GObject introspection.
+ *
+ * Note in addition that unlike ostree_repo_checkout_tree(), the
+ * default is not to use the repository-internal uncompressed objects
+ * cache.
+ */
+gboolean
+ostree_repo_checkout_at (OstreeRepo                        *self,
+                         OstreeRepoCheckoutAtOptions       *options,
+                         int                                destination_dfd,
+                         const char                        *destination_path,
+                         const char                        *commit,
+                         GCancellable                      *cancellable,
+                         GError                           **error)
+{
+  gboolean ret = FALSE;
+  g_autoptr(GFile) commit_root = NULL;
+  g_autoptr(GFile) target_dir = NULL;
+  g_autoptr(GFileInfo) target_info = NULL;
+  OstreeRepoCheckoutAtOptions default_options = { 0, };
+
+  if (!options)
+    {
+      default_options.subpath = NULL;
+      options = &default_options;
+    }
+
+  commit_root = (GFile*) _ostree_repo_file_new_for_commit (self, commit, error);
+  if (!commit_root)
+    goto out;
+
+  if (!ostree_repo_file_ensure_resolved ((OstreeRepoFile*)commit_root, error))
+    goto out;
+
+  if (options->subpath && strcmp (options->subpath, "/") != 0)
+    target_dir = g_file_get_child (commit_root, options->subpath);
+  else
+    target_dir = g_object_ref (commit_root);
+  target_info = g_file_query_info (target_dir, OSTREE_GIO_FAST_QUERYINFO,
+                                   G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+                                   cancellable, error);
+  if (!target_info)
+    goto out;
+
+  if (!checkout_tree_at (self, options,
+                         destination_dfd,
+                         destination_path,
+                         (OstreeRepoFile*)target_dir, target_info,
+                         cancellable, error))
+    goto out;
+
+  ret = TRUE;
+ out:
+  return ret;
+}
+
+static guint
+devino_hash (gconstpointer a)
+{
+  OstreeDevIno *a_i = (gpointer)a;
+  return (guint) (a_i->dev + a_i->ino);
+}
+
+static int
+devino_equal (gconstpointer   a,
+              gconstpointer   b)
+{
+  OstreeDevIno *a_i = (gpointer)a;
+  OstreeDevIno *b_i = (gpointer)b;
+  return a_i->dev == b_i->dev
+    && a_i->ino == b_i->ino;
+}
+
+/**
+ * ostree_repo_devino_cache_new:
+ * 
+ * OSTree has support for pairing ostree_repo_checkout_tree_at() using
+ * hardlinks in combination with a later
+ * ostree_repo_write_directory_to_mtree() using a (normally modified)
+ * directory.  In order for OSTree to optimally detect just the new
+ * files, use this function and fill in the `devino_to_csum_cache`
+ * member of `OstreeRepoCheckoutAtOptions`, then call
+ * ostree_repo_commit_set_devino_cache().
+ *
+ * Returns: (transfer full): Newly allocated cache
+ */
+OstreeRepoDevInoCache *
+ostree_repo_devino_cache_new (void)
+{
+  return (OstreeRepoDevInoCache*) g_hash_table_new_full (devino_hash, devino_equal, g_free, NULL);
+}
+
 /**
  * ostree_repo_checkout_gc:
  * @self: Repo
@@ -724,7 +993,7 @@ ostree_repo_checkout_gc (OstreeRepo        *self,
                          GError           **error)
 {
   gboolean ret = FALSE;
-  gs_unref_hashtable GHashTable *to_clean_dirs = NULL;
+  g_autoptr(GHashTable) to_clean_dirs = NULL;
   GHashTableIter iter;
   gpointer key, value;
 
@@ -737,38 +1006,36 @@ ostree_repo_checkout_gc (OstreeRepo        *self,
     g_hash_table_iter_init (&iter, to_clean_dirs);
   while (to_clean_dirs && g_hash_table_iter_next (&iter, &key, &value))
     {
-      gs_unref_object GFile *objdir = NULL;
-      gs_unref_object GFileEnumerator *enumerator = NULL;
-      gs_free char *objdir_name = NULL;
+      g_autofree char *objdir_name = g_strdup_printf ("%02x", GPOINTER_TO_UINT (key));
+      g_auto(GLnxDirFdIterator) dfd_iter = { 0, };
 
-      objdir_name = g_strdup_printf ("%02x", GPOINTER_TO_UINT (key));
-      objdir = g_file_get_child (self->uncompressed_objects_dir, objdir_name);
-
-      enumerator = g_file_enumerate_children (objdir, "standard::name,standard::type,unix::inode,unix::nlink", 
-                                              G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
-                                              cancellable, 
-                                              error);
-      if (!enumerator)
+      if (!glnx_dirfd_iterator_init_at (self->uncompressed_objects_dir_fd, objdir_name, FALSE,
+                                        &dfd_iter, error))
         goto out;
-  
+
       while (TRUE)
         {
-          GFileInfo *file_info;
-          guint32 nlinks;
+          struct dirent *dent;
+          struct stat stbuf;
 
-          if (!gs_file_enumerator_iterate (enumerator, &file_info, NULL,
-                                           cancellable, error))
+          if (!glnx_dirfd_iterator_next_dent (&dfd_iter, &dent, cancellable, error))
             goto out;
-          if (file_info == NULL)
+          if (dent == NULL)
             break;
-          
-          nlinks = g_file_info_get_attribute_uint32 (file_info, "unix::nlink");
-          if (nlinks == 1)
+
+          if (fstatat (dfd_iter.fd, dent->d_name, &stbuf, AT_SYMLINK_NOFOLLOW) != 0)
             {
-              gs_unref_object GFile *objpath = NULL;
-              objpath = g_file_get_child (objdir, g_file_info_get_name (file_info));
-              if (!gs_file_unlink (objpath, cancellable, error))
-                goto out;
+              glnx_set_error_from_errno (error);
+              goto out;
+            }
+          
+          if (stbuf.st_nlink == 1)
+            {
+              if (unlinkat (dfd_iter.fd, dent->d_name, 0) != 0)
+                {
+                  glnx_set_error_from_errno (error);
+                  goto out;
+                }
             }
         }
     }
