@@ -21,12 +21,11 @@
 #include "config.h"
 
 #include "otutil.h"
-#include "libgsystem.h"
 
 #include "ostree-sysroot-upgrader.h"
 
 /**
- * SECTION:libostree-sysroot-upgrader
+ * SECTION:ostree-sysroot-upgrader
  * @title: Simple upgrade class
  * @short_description: Upgrade OSTree systems
  *
@@ -42,11 +41,13 @@ struct OstreeSysrootUpgrader {
 
   OstreeSysroot *sysroot;
   char *osname;
+  OstreeSysrootUpgraderFlags flags;
 
   OstreeDeployment *merge_deployment;
   GKeyFile *origin;
   char *origin_remote;
   char *origin_ref;
+  char *override_csum;
 
   char *new_revision;
 }; 
@@ -55,7 +56,8 @@ enum {
   PROP_0,
 
   PROP_SYSROOT,
-  PROP_OSNAME
+  PROP_OSNAME,
+  PROP_FLAGS
 };
 
 static void ostree_sysroot_upgrader_initable_iface_init (GInitableIface *iface);
@@ -69,7 +71,21 @@ parse_refspec (OstreeSysrootUpgrader  *self,
                GError                **error)
 {
   gboolean ret = FALSE;
-  gs_free char *origin_refspec = NULL;
+  g_autofree char *origin_refspec = NULL;
+  g_autofree char *unconfigured_state = NULL;
+  g_autofree char *csum = NULL;
+
+  if ((self->flags & OSTREE_SYSROOT_UPGRADER_FLAGS_IGNORE_UNCONFIGURED) == 0)
+    {
+      /* If explicit action by the OS creator is requried to upgrade, print their text as an error */
+      unconfigured_state = g_key_file_get_string (self->origin, "origin", "unconfigured-state", NULL);
+      if (unconfigured_state)
+        {
+          g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                       "origin unconfigured-state: %s", unconfigured_state);
+          goto out;
+        }
+    }
 
   origin_refspec = g_key_file_get_string (self->origin, "origin", "refspec", NULL);
   if (!origin_refspec)
@@ -85,7 +101,12 @@ parse_refspec (OstreeSysrootUpgrader  *self,
                              &self->origin_ref,
                              error))
     goto out;
-  
+
+  csum = g_key_file_get_string (self->origin, "origin", "override-commit", NULL);
+  if (csum != NULL && !ostree_validate_checksum_string (csum, error))
+    goto out;
+  self->override_csum = g_steal_pointer (&csum);
+
   ret = TRUE;
  out:
   return ret;
@@ -100,8 +121,6 @@ ostree_sysroot_upgrader_initable_init (GInitable        *initable,
   OstreeSysrootUpgrader *self = (OstreeSysrootUpgrader*)initable;
   OstreeDeployment *booted_deployment =
     ostree_sysroot_get_booted_deployment (self->sysroot);
-  gs_unref_object GFile *deployment_path = NULL;
-  gs_unref_object GFile *deployment_origin_path = NULL;
 
   if (booted_deployment == NULL && self->osname == NULL)
     {
@@ -130,9 +149,6 @@ ostree_sysroot_upgrader_initable_init (GInitable        *initable,
       goto out;
     }
 
-  deployment_path = ostree_sysroot_get_deployment_directory (self->sysroot, self->merge_deployment);
-  deployment_origin_path = ostree_sysroot_get_deployment_origin_path (deployment_path);
-
   self->origin = ostree_deployment_get_origin (self->merge_deployment);
   if (!self->origin)
     {
@@ -142,6 +158,7 @@ ostree_sysroot_upgrader_initable_init (GInitable        *initable,
                    ostree_deployment_get_deployserial (self->merge_deployment));
       goto out;
     }
+  g_key_file_ref (self->origin);
 
   if (!parse_refspec (self, cancellable, error))
     goto out;
@@ -170,6 +187,7 @@ ostree_sysroot_upgrader_finalize (GObject *object)
     g_key_file_unref (self->origin);
   g_free (self->origin_remote);
   g_free (self->origin_ref);
+  g_free (self->override_csum);
 
   G_OBJECT_CLASS (ostree_sysroot_upgrader_parent_class)->finalize (object);
 }
@@ -190,6 +208,9 @@ ostree_sysroot_upgrader_set_property (GObject         *object,
     case PROP_OSNAME:
       self->osname = g_value_dup_string (value);
       break;
+    case PROP_FLAGS:
+      self->flags = g_value_get_flags (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -212,6 +233,9 @@ ostree_sysroot_upgrader_get_property (GObject         *object,
     case PROP_OSNAME:
       g_value_set_string (value, self->osname);
       break;
+    case PROP_FLAGS:
+      g_value_set_flags (value, self->flags);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -248,6 +272,13 @@ ostree_sysroot_upgrader_class_init (OstreeSysrootUpgraderClass *klass)
                                    PROP_OSNAME,
                                    g_param_spec_string ("osname", "", "", NULL,
                                                         G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+
+  g_object_class_install_property (object_class,
+                                   PROP_FLAGS,
+                                   g_param_spec_flags ("flags", "", "",
+                                                       ostree_sysroot_upgrader_flags_get_type (),
+                                                       0,
+                                                       G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 }
 
 static void
@@ -287,6 +318,25 @@ ostree_sysroot_upgrader_new_for_os (OstreeSysroot *sysroot,
                        "sysroot", sysroot, "osname", osname, NULL);
 }
 
+/**
+ * ostree_sysroot_upgrader_new_for_os_with_flags:
+ * @sysroot: An #OstreeSysroot
+ * @osname: (allow-none): Operating system name
+ * @flags: Flags
+ *
+ * Returns: (transfer full): An upgrader
+ */
+OstreeSysrootUpgrader *
+ostree_sysroot_upgrader_new_for_os_with_flags (OstreeSysroot              *sysroot,
+                                               const char                 *osname,
+                                               OstreeSysrootUpgraderFlags  flags,
+                                               GCancellable               *cancellable,
+                                               GError                    **error)
+{
+  return g_initable_new (OSTREE_TYPE_SYSROOT_UPGRADER, cancellable, error,
+                         "sysroot", sysroot, "osname", osname, "flags", flags, NULL);
+}
+
 /**
  * ostree_sysroot_upgrader_get_origin:
  * @self: Sysroot
@@ -299,6 +349,33 @@ ostree_sysroot_upgrader_get_origin (OstreeSysrootUpgrader *self)
   return self->origin;
 }
 
+/**
+ * ostree_sysroot_upgrader_dup_origin:
+ * @self: Sysroot
+ *
+ * Returns: (transfer full): A copy of the origin file, or %NULL if unknown
+ */
+GKeyFile *
+ostree_sysroot_upgrader_dup_origin (OstreeSysrootUpgrader *self)
+{
+  GKeyFile *copy = NULL;
+
+  g_return_val_if_fail (OSTREE_IS_SYSROOT_UPGRADER (self), NULL);
+
+  if (self->origin != NULL)
+    {
+      g_autofree char *data = NULL;
+      gsize length = 0;
+
+      copy = g_key_file_new ();
+      data = g_key_file_to_data (self->origin, &length, NULL);
+      g_key_file_load_from_data (copy, data, length,
+                                 G_KEY_FILE_KEEP_COMMENTS, NULL);
+    }
+
+  return copy;
+}
+
 /**
  * ostree_sysroot_upgrader_set_origin:
  * @self: Sysroot
@@ -361,8 +438,8 @@ ostree_sysroot_upgrader_check_timestamps (OstreeRepo     *repo,
                                           GError        **error)
 {
   gboolean ret = FALSE;
-  gs_unref_variant GVariant *old_commit = NULL;
-  gs_unref_variant GVariant *new_commit = NULL;
+  g_autoptr(GVariant) old_commit = NULL;
+  g_autoptr(GVariant) new_commit = NULL;
 
   if (!ostree_repo_load_variant (repo,
                                  OSTREE_OBJECT_TYPE_COMMIT,
@@ -380,8 +457,8 @@ ostree_sysroot_upgrader_check_timestamps (OstreeRepo     *repo,
     {
       GDateTime *old_ts = g_date_time_new_from_unix_utc (ostree_commit_get_timestamp (old_commit));
       GDateTime *new_ts = g_date_time_new_from_unix_utc (ostree_commit_get_timestamp (new_commit));
-      gs_free char *old_ts_str = NULL;
-      gs_free char *new_ts_str = NULL;
+      g_autofree char *old_ts_str = NULL;
+      g_autofree char *new_ts_str = NULL;
 
       g_assert (old_ts);
       g_assert (new_ts);
@@ -427,13 +504,39 @@ ostree_sysroot_upgrader_pull (OstreeSysrootUpgrader  *self,
                               gboolean               *out_changed,
                               GCancellable           *cancellable,
                               GError                **error)
+{
+  return ostree_sysroot_upgrader_pull_one_dir (self, NULL, flags, upgrader_flags, progress, out_changed, cancellable, error);
+}
+
+/**
+ * ostree_sysroot_upgrader_pull_one_dir:
+ *
+ * Like ostree_sysroot_upgrader_pull(), but allows retrieving just a
+ * subpath of the tree.  This can be used to download metadata files
+ * from inside the tree such as package databases.
+ *
+ */
+gboolean
+ostree_sysroot_upgrader_pull_one_dir (OstreeSysrootUpgrader  *self,
+                                      const char             *dir_to_pull,
+                                      OstreeRepoPullFlags     flags,
+                                      OstreeSysrootUpgraderPullFlags     upgrader_flags,
+                                      OstreeAsyncProgress    *progress,
+                                      gboolean               *out_changed,
+                                      GCancellable           *cancellable,
+                                      GError                **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object OstreeRepo *repo = NULL;
-  char *refs_to_fetch[] = { self->origin_ref, NULL };
+  glnx_unref_object OstreeRepo *repo = NULL;
+  char *refs_to_fetch[] = { NULL, NULL };
   const char *from_revision = NULL;
-  gs_free char *new_revision = NULL;
-  gs_free char *origin_refspec = NULL;
+  g_autofree char *new_revision = NULL;
+  g_autofree char *origin_refspec = NULL;
+
+  if (self->override_csum != NULL)
+    refs_to_fetch[0] = self->override_csum;
+  else
+    refs_to_fetch[0] = self->origin_ref;
 
   if (!ostree_sysroot_get_repo (self->sysroot, &repo, cancellable, error))
     goto out;
@@ -448,15 +551,34 @@ ostree_sysroot_upgrader_pull (OstreeSysrootUpgrader  *self,
 
   if (self->origin_remote)
     {
-      if (!ostree_repo_pull (repo, self->origin_remote, refs_to_fetch,
+      if (!ostree_repo_pull_one_dir (repo, self->origin_remote, dir_to_pull, refs_to_fetch,
                              flags, progress,
                              cancellable, error))
         goto out;
+
+      if (progress)
+        ostree_async_progress_finish (progress);
     }
 
-  if (!ostree_repo_resolve_rev (repo, origin_refspec, FALSE, &self->new_revision,
-                                error))
-    goto out;
+  if (self->override_csum != NULL)
+    {
+      if (!ostree_repo_set_ref_immediate (repo,
+                                          self->origin_remote,
+                                          self->origin_ref,
+                                          self->override_csum,
+                                          cancellable,
+                                          error))
+        goto out;
+
+      self->new_revision = g_strdup (self->override_csum);
+    }
+  else
+    {
+      if (!ostree_repo_resolve_rev (repo, origin_refspec, FALSE,
+                                    &self->new_revision, error))
+        goto out;
+
+    }
 
   if (g_strcmp0 (from_revision, self->new_revision) == 0)
     {
@@ -497,7 +619,7 @@ ostree_sysroot_upgrader_deploy (OstreeSysrootUpgrader  *self,
                                 GError                **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object OstreeDeployment *new_deployment = NULL;
+  glnx_unref_object OstreeDeployment *new_deployment = NULL;
 
   if (!ostree_sysroot_deploy_tree (self->sysroot, self->osname,
                                    self->new_revision,
@@ -519,3 +641,22 @@ ostree_sysroot_upgrader_deploy (OstreeSysrootUpgrader  *self,
  out:
   return ret;
 }
+
+GType
+ostree_sysroot_upgrader_flags_get_type (void)
+{
+  static volatile gsize g_define_type_id__volatile = 0;
+
+  if (g_once_init_enter (&g_define_type_id__volatile))
+    {
+      static const GFlagsValue values[] = {
+        { OSTREE_SYSROOT_UPGRADER_FLAGS_IGNORE_UNCONFIGURED, "OSTREE_SYSROOT_UPGRADER_FLAGS_IGNORE_UNCONFIGURED", "ignore-unconfigured" },
+        { 0, NULL, NULL }
+      };
+      GType g_define_type_id =
+        g_flags_register_static (g_intern_static_string ("OstreeSysrootUpgraderFlags"), values);
+      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
+    }
+
+  return g_define_type_id__volatile;
+}
