@@ -56,8 +56,8 @@ ot_gfile_from_build_path (const char *first, ...)
 {
   va_list args;
   const char *arg;
-  gs_free char *path = NULL;
-  gs_unref_ptrarray GPtrArray *components = NULL;  
+  g_autofree char *path = NULL;
+  g_autoptr(GPtrArray) components = NULL;  
 
   va_start (args, first);
 
@@ -98,6 +98,8 @@ ot_gfile_get_child_strconcat (GFile *parent,
   while ((arg = va_arg (args, const char *)) != NULL)
     g_string_append (buf, arg);
 
+  va_end (args);
+
   ret = g_file_get_child (parent, buf->str);
   
   g_string_free (buf, TRUE);
@@ -111,8 +113,8 @@ ot_gfile_get_child_build_path (GFile      *parent,
 {
   va_list args;
   const char *arg;
-  gs_free char *path = NULL;
-  gs_unref_ptrarray GPtrArray *components = NULL;  
+  g_autofree char *path = NULL;
+  g_autoptr(GPtrArray) components = NULL;  
 
   va_start (args, first);
 
@@ -140,7 +142,7 @@ ot_gfile_resolve_path_printf (GFile       *path,
                               ...)
 {
   va_list args;
-  gs_free char *relpath = NULL;
+  g_autofree char *relpath = NULL;
 
   va_start (args, format);
   relpath = g_strdup_vprintf (format, args);
@@ -159,8 +161,8 @@ ot_gfile_get_symlink_target_from_info (GFile             *path,
 {
   gboolean ret = FALSE;
   const char *target;
-  gs_unref_object GFile *path_parent = NULL;
-  gs_unref_object GFile *ret_target = NULL;
+  g_autoptr(GFile) path_parent = NULL;
+  g_autoptr(GFile) ret_target = NULL;
 
   if (g_file_info_get_file_type (file_info) != G_FILE_TYPE_SYMBOLIC_LINK)
     {
@@ -189,7 +191,7 @@ ot_gfile_query_info_allow_noent (GFile                *path,
                                  GError              **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object GFileInfo *ret_file_info = NULL;
+  g_autoptr(GFileInfo) ret_file_info = NULL;
   GError *temp_error = NULL;
 
   ret_file_info = g_file_query_info (path, queryopts, flags,
@@ -220,8 +222,8 @@ ot_gfile_query_symlink_target_allow_noent (GFile          *path,
                                            GError        **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object GFileInfo *file_info = NULL;
-  gs_unref_object GFile *ret_target = NULL;
+  g_autoptr(GFileInfo) file_info = NULL;
+  g_autoptr(GFile) ret_target = NULL;
 
   if (!ot_gfile_query_info_allow_noent (path, OSTREE_GIO_FAST_QUERYINFO,
                                         G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
@@ -250,9 +252,10 @@ ot_gfile_load_contents_utf8_allow_noent (GFile          *path,
 {
   gboolean ret = FALSE;
   GError *temp_error = NULL;
-  gs_free char *ret_contents = NULL;
+  g_autofree char *ret_contents = NULL;
 
-  ret_contents = gs_file_load_contents_utf8 (path, cancellable, &temp_error);
+  ret_contents = glnx_file_get_contents_utf8_at (AT_FDCWD, gs_file_get_path_cached (path), NULL,
+                                                 cancellable, &temp_error);
   if (!ret_contents)
     {
       if (g_error_matches (temp_error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
@@ -283,65 +286,13 @@ ot_gfile_replace_contents_fsync (GFile          *path,
                                  GCancellable   *cancellable,
                                  GError        **error)
 {
-  gboolean ret = FALSE;
-  int parent_dfd;
-  int fd;
-  const char *target_basename = gs_file_get_basename_cached (path);
-  gs_unref_object GFile *parent = NULL;
-  gs_free char *tmpname = NULL;
-  gs_unref_object GOutputStream *stream = NULL;
-  gs_unref_object GInputStream *instream = NULL;
+  gsize len;
+  const guint8*buf = g_bytes_get_data (contents, &len);
 
-  parent = g_file_get_parent (path);
-
-  if (!gs_file_open_dir_fd (parent, &parent_dfd,
-                            cancellable, error))
-    goto out;
-
-  if (!gs_file_open_in_tmpdir_at (parent_dfd, 0644,
-                                  &tmpname, &stream,
-                                  cancellable, error))
-    goto out;
-
-  g_assert (G_IS_FILE_DESCRIPTOR_BASED (stream));
-  fd = g_file_descriptor_based_get_fd (G_FILE_DESCRIPTOR_BASED (stream));
-
-  instream = g_memory_input_stream_new_from_bytes (contents);
-
-  if (posix_fallocate (fd, 0, g_bytes_get_size (contents)) != 0)
-    {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
-    }
-
-  if (g_output_stream_splice (stream, instream, 0,
-                              cancellable, error) < 0)
-    goto out;
-
-  if (fdatasync (fd) != 0)
-    {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
-    }
-
-  if (!g_output_stream_close (stream, cancellable, error))
-    goto out;
-
-  if (renameat (parent_dfd, tmpname, parent_dfd, target_basename) == -1)
-    {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
-    }
-
-  g_clear_pointer (&tmpname, g_free);
-
-  ret = TRUE;
- out:
-  if (tmpname)
-    (void) unlinkat (parent_dfd, tmpname, 0);
-  if (parent_dfd != -1)
-    (void) close (parent_dfd);
-  return ret;
+  return glnx_file_replace_contents_at (AT_FDCWD, gs_file_get_path_cached (path),
+                                        buf, len,
+                                        GLNX_FILE_REPLACE_DATASYNC_NEW,
+                                        cancellable, error);
 }
 
 /**
@@ -355,25 +306,12 @@ ot_gfile_ensure_unlinked (GFile         *path,
                           GCancellable  *cancellable,
                           GError       **error)
 {
-  gboolean ret = FALSE;
-  GError *temp_error = NULL;
-
-  if (!gs_file_unlink (path, cancellable, &temp_error))
+  if (unlink (gs_file_get_path_cached (path)) != 0)
     {
-      if (g_error_matches (temp_error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
-        {
-          g_clear_error (&temp_error);
-        }
-      else
-        {
-          g_propagate_error (error, temp_error);
-          goto out;
-        }
+      if (errno != ENOENT)
+        return FALSE;
     }
-  
-  ret = TRUE;
- out:
-  return ret;
+  return TRUE;
 }
 
 /**
@@ -390,21 +328,20 @@ ot_util_fsync_directory (GFile         *dir,
                          GError       **error)
 {
   gboolean ret = FALSE;
-  int dfd = -1;
+  glnx_fd_close int dfd = -1;
 
-  if (!gs_file_open_dir_fd (dir, &dfd, cancellable, error))
+  if (!glnx_opendirat (AT_FDCWD, gs_file_get_path_cached (dir), TRUE,
+                       &dfd, error))
     goto out;
 
   if (fsync (dfd) != 0)
     {
-      ot_util_set_error_from_errno (error, errno);
+      glnx_set_error_from_errno (error);
       goto out;
     }
 
   ret = TRUE;
  out:
-  if (dfd != -1)
-    (void) close (dfd);
   return ret;
 }
 
@@ -423,9 +360,9 @@ ot_util_ensure_directory_and_fsync (GFile         *dir,
                                     GError       **error)
 {
   gboolean ret = FALSE;
-  int parentfd = -1;
-  const char *basename = gs_file_get_basename_cached (dir);
-  gs_unref_object GFile *parent = g_file_get_parent (dir);
+  glnx_fd_close int parentfd = -1;
+  const char *basename = glnx_basename (gs_file_get_path_cached (dir));
+  g_autoptr(GFile) parent = g_file_get_parent (dir);
   
  again:
   parentfd = open (gs_file_get_path_cached (parent),
@@ -472,63 +409,93 @@ ot_util_ensure_directory_and_fsync (GFile         *dir,
 
   ret = TRUE;
  out:
-  if (parentfd != -1)
-    (void) close (parentfd);
   return ret;
 }
 
-/**
- * ot_gfile_atomic_symlink_swap:
- * @path: Replace the contents of this symbolic link
- * @target: New symbolic link target
- * @cancellable:
- * @error
- *
- * Create a new temporary symbolic link, then use the Unix rename()
- * function to atomically replace @path with the new symbolic link.
- * Do not use this function inside directories such as /tmp as it uses
- * a predicatable file name.
- */
+#if !GLIB_CHECK_VERSION(2, 44, 0)
+
 gboolean
-ot_gfile_atomic_symlink_swap (GFile          *path,
-                              const char     *target,
-                              GCancellable   *cancellable,
-                              GError        **error)
+ot_file_enumerator_iterate (GFileEnumerator  *direnum,
+                            GFileInfo       **out_info,
+                            GFile           **out_child,
+                            GCancellable     *cancellable,
+                            GError          **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object GFile *parent = g_file_get_parent (path);
-  gs_free char *tmpname = g_strconcat (gs_file_get_basename_cached (path), ".tmp", NULL);
-  gs_unref_object GFile *tmppath = g_file_get_child (parent, tmpname);
-  int parent_dfd = -1;
+  GError *temp_error = NULL;
 
-  if (!ot_gfile_ensure_unlinked (tmppath, cancellable, error))
-    goto out;
-  
-  if (!g_file_make_symbolic_link (tmppath, target, cancellable, error))
-    goto out;
+  static GQuark cached_info_quark;
+  static GQuark cached_child_quark;
+  static gsize quarks_initialized;
 
-  if (!gs_file_open_dir_fd (parent, &parent_dfd, cancellable, error))
-    goto out;
+  g_return_val_if_fail (direnum != NULL, FALSE);
+  g_return_val_if_fail (out_info != NULL, FALSE);
 
-  /* Ensure the link has hit disk */
-  if (fsync (parent_dfd) != 0)
+  if (g_once_init_enter (&quarks_initialized))
     {
-      ot_util_set_error_from_errno (error, errno);
-      goto out;
+      cached_info_quark = g_quark_from_static_string ("ot-cached-info");
+      cached_child_quark = g_quark_from_static_string ("ot-cached-child");
+      g_once_init_leave (&quarks_initialized, 1);
     }
 
-  if (!gs_file_rename (tmppath, path, cancellable, error))
-    goto out;
-
-  /* And sync again for good measure */
-  if (fsync (parent_dfd) != 0)
+  *out_info = g_file_enumerator_next_file (direnum, cancellable, &temp_error);
+  if (out_child)
+    *out_child = NULL;
+  if (temp_error != NULL)
     {
-      ot_util_set_error_from_errno (error, errno);
+      g_propagate_error (error, temp_error);
       goto out;
     }
+  else if (*out_info != NULL)
+    {
+      g_object_set_qdata_full ((GObject*)direnum, cached_info_quark, *out_info, (GDestroyNotify)g_object_unref);
+      if (out_child != NULL)
+        {
+          const char *name = g_file_info_get_name (*out_info);
+          *out_child = g_file_get_child (g_file_enumerator_get_container (direnum), name);
+          g_object_set_qdata_full ((GObject*)direnum, cached_child_quark, *out_child, (GDestroyNotify)g_object_unref);
+        }
+    }
 
   ret = TRUE;
  out:
-  if (parent_dfd != -1) (void) close (parent_dfd);
   return ret;
 }
+
+#endif
+
+G_LOCK_DEFINE_STATIC (pathname_cache);
+
+/**
+ * ot_file_get_path_cached:
+ *
+ * Like g_file_get_path(), but returns a constant copy so callers
+ * don't need to free the result.
+ */
+const char *
+ot_file_get_path_cached (GFile *file)
+{
+  const char *path;
+  static GQuark _file_path_quark = 0;
+
+  if (G_UNLIKELY (_file_path_quark) == 0)
+    _file_path_quark = g_quark_from_static_string ("gsystem-file-path");
+
+  G_LOCK (pathname_cache);
+
+  path = g_object_get_qdata ((GObject*)file, _file_path_quark);
+  if (!path)
+    {
+      path = g_file_get_path (file);
+      if (path == NULL)
+        {
+          G_UNLOCK (pathname_cache);
+          return NULL;
+        }
+      g_object_set_qdata_full ((GObject*)file, _file_path_quark, (char*)path, (GDestroyNotify)g_free);
+    }
+
+  G_UNLOCK (pathname_cache);
+
+  return path;
+}
