@@ -69,6 +69,13 @@ gboolean ot_gfile_load_contents_utf8_allow_noent (GFile          *path,
                                                   GCancellable   *cancellable,
                                                   GError        **error);
 
+gboolean ot_file_replace_contents_at (int             dfd,
+                                      const char     *path,
+                                      GBytes         *contents,
+                                      gboolean        datasync,
+                                      GCancellable   *cancellable,
+                                      GError        **error);
+
 gboolean ot_gfile_replace_contents_fsync (GFile          *path,
                                           GBytes         *contents,
                                           GCancellable   *cancellable,
@@ -78,11 +85,6 @@ gboolean ot_gfile_ensure_unlinked (GFile         *path,
                                    GCancellable  *cancellable,
                                    GError       **error);
 
-gboolean ot_gfile_atomic_symlink_swap (GFile          *path,
-                                       const char     *target,
-                                       GCancellable   *cancellable,
-                                       GError        **error);
-
 gboolean ot_util_ensure_directory_and_fsync (GFile         *dir,
                                              GCancellable  *cancellable,
                                              GError       **error);
@@ -91,5 +93,36 @@ gboolean ot_util_fsync_directory (GFile         *dir,
                                   GCancellable  *cancellable,
                                   GError       **error);
 
-G_END_DECLS
+#if !GLIB_CHECK_VERSION(2, 44, 0)
+gboolean
+ot_file_enumerator_iterate (GFileEnumerator  *direnum,
+                            GFileInfo       **out_info,
+                            GFile           **out_child,
+                            GCancellable     *cancellable,
+                            GError          **error);
+#else
+static inline gboolean
+ot_file_enumerator_iterate (GFileEnumerator  *direnum,
+                            GFileInfo       **out_info,
+                            GFile           **out_child,
+                            GCancellable     *cancellable,
+                            GError          **error)
+{
+  G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
+  return (g_file_enumerator_iterate) (direnum, out_info, out_child, cancellable, error);
+  G_GNUC_END_IGNORE_DEPRECATIONS;
+}
+#endif
+#define g_file_enumerator_iterate ot_file_enumerator_iterate
 
+const char *
+ot_file_get_path_cached (GFile *file);
+
+static inline
+const char *
+gs_file_get_path_cached (GFile *file)
+{
+  return ot_file_get_path_cached (file);
+}
+
+G_END_DECLS
