@@ -30,6 +30,15 @@
 
 #include "otutil.h"
 
+GVariant *
+ot_gvariant_new_empty_string_dict (void)
+{
+  g_auto(GVariantBuilder) builder = {{0,}};
+
+  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sv}"));
+  return g_variant_builder_end (&builder);
+}
+
 GVariant *
 ot_gvariant_new_bytearray (const guchar   *data,
                            gsize           len)
@@ -79,7 +88,7 @@ ot_util_variant_save (GFile *dest,
                       GError  **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object GOutputStream *out = NULL;
+  g_autoptr(GOutputStream) out = NULL;
   gsize bytes_written;
   
   out = (GOutputStream*)g_file_replace (dest, NULL, FALSE, G_FILE_CREATE_REPLACE_DESTINATION,
@@ -108,47 +117,34 @@ ot_util_variant_take_ref (GVariant *variant)
   return g_variant_take_ref (variant);
 }
 
-/**
- * ot_util_variant_map:
- * @src: a #GFile
- * @type: Use this for variant
- * @trusted: See documentation of g_variant_new_from_data()
- * @out_variant: (out): Return location for new variant
- * @error:
- *
- * Memory-map @src, and store a new #GVariant referring to this memory
- * in @out_variant.  Note the returned @out_variant is not floating.
- */
 gboolean
-ot_util_variant_map (GFile              *src,
-                     const GVariantType *type,
-                     gboolean            trusted,
-                     GVariant          **out_variant,
-                     GError            **error)
+ot_util_variant_map_at (int dfd,
+                        const char *path,
+                        const GVariantType *type,
+                        OtVariantMapFlags flags,
+                        GVariant **out_variant,
+                        GError  **error)
 {
-  gboolean ret = FALSE;
-  gs_unref_variant GVariant *ret_variant = NULL;
-  GMappedFile *mfile = NULL;
+  glnx_fd_close int fd = -1;
+  const gboolean trusted = (flags & OT_VARIANT_MAP_TRUSTED) > 0;
 
-  mfile = gs_file_map_noatime (src, NULL, error);
-  if (!mfile)
-    goto out;
+  fd = openat (dfd, path, O_RDONLY | O_CLOEXEC);
+  if (fd < 0)
+    {
+      if (errno == ENOENT && (flags & OT_VARIANT_MAP_ALLOW_NOENT) > 0)
+        {
+          *out_variant = NULL;
+          return TRUE;
+        }
+      else
+        {
+          glnx_set_error_from_errno (error);
+          g_prefix_error (error, "Opening %s: ", path);
+          return FALSE;
+        }
+    }
 
-  ret_variant = g_variant_new_from_data (type,
-                                         g_mapped_file_get_contents (mfile),
-                                         g_mapped_file_get_length (mfile),
-                                         trusted,
-                                         (GDestroyNotify) g_mapped_file_unref,
-                                         mfile);
-  mfile = NULL;
-  g_variant_ref_sink (ret_variant);
-  
-  ret = TRUE;
-  ot_transfer_out_value(out_variant, &ret_variant);
- out:
-  if (mfile)
-    g_mapped_file_unref (mfile);
-  return ret;
+  return ot_util_variant_map_fd (fd, 0, type, trusted, out_variant, error);
 }
 
 typedef struct {
@@ -161,10 +157,11 @@ variant_map_data_destroy (gpointer data)
 {
   VariantMapData *mdata = data;
   (void) munmap (mdata->addr, mdata->len);
+  g_free (mdata);
 }
 
 gboolean
-ot_util_variant_map_fd (GFileDescriptorBased  *stream,
+ot_util_variant_map_fd (int                    fd,
                         goffset                start,
                         const GVariantType    *type,
                         gboolean               trusted,
@@ -177,15 +174,17 @@ ot_util_variant_map_fd (GFileDescriptorBased  *stream,
   VariantMapData *mdata = NULL;
   gsize len;
 
-  if (!gs_stream_fstat (stream, &stbuf, NULL, error))
-    goto out;
+  if (fstat (fd, &stbuf) != 0)
+    {
+      glnx_set_error_from_errno (error);
+      goto out;
+    }
 
   len = stbuf.st_size - start;
-  map = mmap (NULL, len, PROT_READ, MAP_PRIVATE, 
-              g_file_descriptor_based_get_fd (stream), start);
+  map = mmap (NULL, len, PROT_READ, MAP_PRIVATE, fd, start);
   if (!map)
     {
-      ot_util_set_error_from_errno (error, errno);
+      glnx_set_error_from_errno (error);
       goto out;
     }
 
@@ -194,8 +193,8 @@ ot_util_variant_map_fd (GFileDescriptorBased  *stream,
   mdata->len = len;
 
   ret = TRUE;
-  *out_variant = g_variant_new_from_data (type, map, len, trusted,
-                                          variant_map_data_destroy, mdata);
+  *out_variant = g_variant_ref_sink (g_variant_new_from_data (type, map, len, trusted,
+                                                              variant_map_data_destroy, mdata));
  out:
   return ret;
 }
@@ -215,8 +214,8 @@ ot_util_variant_from_stream (GInputStream         *src,
                              GError              **error)
 {
   gboolean ret = FALSE;
-  gs_unref_object GMemoryOutputStream *data_stream = NULL;
-  gs_unref_variant GVariant *ret_variant = NULL;
+  g_autoptr(GMemoryOutputStream) data_stream = NULL;
+  g_autoptr(GVariant) ret_variant = NULL;
 
   data_stream = (GMemoryOutputStream*)g_memory_output_stream_new (NULL, 0, g_realloc, g_free);
 
@@ -255,32 +254,79 @@ ot_util_variant_builder_from_variant (GVariant            *variant,
                                       const GVariantType  *type)
 {
   GVariantBuilder *builder = NULL;
-  gint i, n;
   
   builder = g_variant_builder_new (type);
   
-  n = g_variant_n_children (variant);
-  for (i = 0; i < n; i++)
+  if (variant != NULL)
     {
-      GVariant *child = g_variant_get_child_value (variant, i);
-      g_variant_builder_add_value (builder, child);
-      g_variant_unref (child);
+      gint i, n;
+
+      n = g_variant_n_children (variant);
+      for (i = 0; i < n; i++)
+        {
+          GVariant *child = g_variant_get_child_value (variant, i);
+          g_variant_builder_add_value (builder, child);
+          g_variant_unref (child);
+        }
     }
     
   return builder;
 }
 
-GVariant *
-ot_variant_new_from_bytes (const GVariantType  *type,
-                           GBytes        *bytes,
-                           gboolean       trusted)
+/**
+ * ot_variant_bsearch_str:
+ * @array: A GVariant array whose first element must be a string
+ * @str: Search for this string
+ * @out_pos: Output position
+ *
+ *
+ * Binary search in a GVariant array, which must be of the form 'a(s...)',
+ * where '...' may be anything.  The array elements must be sorted.
+ *
+ * Returns: %TRUE if found, %FALSE otherwise
+ */
+gboolean
+ot_variant_bsearch_str (GVariant   *array,
+                        const char *str,
+                        int        *out_pos)
 {
-#if GLIB_VERSION_MIN_REQUIRED >= GLIB_VERSION_2_36
-  return g_variant_new_from_bytes (type, bytes, trusted);
-#else
-  gsize size;
-  gconstpointer data = g_bytes_get_data (bytes, &size);
-  return g_variant_new_from_data (type, data, size, trusted,
-                                  (GDestroyNotify)g_bytes_unref, bytes);
-#endif
+  gsize imax, imin;
+  gsize imid = -1;
+  gsize n;
+
+  n = g_variant_n_children (array);
+  if (n == 0)
+    return FALSE;
+
+  imax = n - 1;
+  imin = 0;
+  while (imax >= imin)
+    {
+      g_autoptr(GVariant) child = NULL;
+      const char *cur;
+      int cmp;
+
+      imid = (imin + imax) / 2;
+
+      child = g_variant_get_child_value (array, imid);
+      g_variant_get_child (child, 0, "&s", &cur, NULL);
+
+      cmp = strcmp (cur, str);
+      if (cmp < 0)
+        imin = imid + 1;
+      else if (cmp > 0)
+        {
+          if (imid == 0)
+            break;
+          imax = imid - 1;
+        }
+      else
+        {
+          *out_pos = imid;
+          return TRUE;
+        }
+    }
+
+  *out_pos = imid;
+  return FALSE;
 }
