@@ -22,10 +22,10 @@
 
 #include "config.h"
 
+#include "ot-main.h"
 #include "ot-admin-builtins.h"
 #include "ot-admin-functions.h"
 #include "otutil.h"
-#include "libgsystem.h"
 
 #include <glib/gi18n.h>
 
@@ -34,20 +34,21 @@ static GOptionEntry options[] = {
 };
 
 gboolean
-ot_admin_builtin_init_fs (int argc, char **argv, OstreeSysroot *sysroot, GCancellable *cancellable, GError **error)
+ot_admin_builtin_init_fs (int argc, char **argv, GCancellable *cancellable, GError **error)
 {
   GOptionContext *context;
+  glnx_unref_object OstreeSysroot *sysroot = NULL;
   gboolean ret = FALSE;
-  gs_unref_object GFile *dir = NULL;
-  gs_unref_object GFile *child = NULL;
-  gs_unref_object OstreeSysroot *target_sysroot = NULL;
+  glnx_fd_close int root_dfd = -1;
+  glnx_unref_object OstreeSysroot *target_sysroot = NULL;
   guint i;
   const char *normal_toplevels[] = {"boot", "dev", "home", "proc", "run", "sys"};
 
   context = g_option_context_new ("PATH - Initialize a root filesystem");
-  g_option_context_add_main_entries (context, options, NULL);
 
-  if (!g_option_context_parse (context, &argc, &argv, error))
+  if (!ostree_admin_option_context_parse (context, options, &argc, &argv,
+                                          OSTREE_ADMIN_BUILTIN_FLAG_SUPERUSER | OSTREE_ADMIN_BUILTIN_FLAG_UNLOCKED,
+                                          &sysroot, cancellable, error))
     goto out;
 
   if (argc < 2)
@@ -56,26 +57,31 @@ ot_admin_builtin_init_fs (int argc, char **argv, OstreeSysroot *sysroot, GCancel
       goto out;
     }
 
-  dir = g_file_new_for_path (argv[1]);
-  target_sysroot = ostree_sysroot_new (dir);
+  if (!glnx_opendirat (AT_FDCWD, argv[1], TRUE, &root_dfd, error))
+    goto out;
+  { g_autoptr(GFile) dir = g_file_new_for_path (argv[1]);
+    target_sysroot = ostree_sysroot_new (dir);
+  }
 
   for (i = 0; i < G_N_ELEMENTS(normal_toplevels); i++)
     {
-      child = g_file_get_child (dir, normal_toplevels[i]);
-      if (!gs_file_ensure_directory_mode (child, 0755, cancellable, error))
+      if (!glnx_shutil_mkdir_p_at (root_dfd, normal_toplevels[i], 0755,
+                                   cancellable, error))
         goto out;
-      g_clear_object (&child);
     }
-
-  child = g_file_get_child (dir, "root");
-  if (!gs_file_ensure_directory_mode (child, 0700, cancellable, error))
+  
+  if (!glnx_shutil_mkdir_p_at (root_dfd, "root", 0700,
+                               cancellable, error))
     goto out;
-  g_clear_object (&child);
 
-  child = g_file_get_child (dir, "tmp");
-  if (!gs_file_ensure_directory_mode (child, 01777, cancellable, error))
+  if (!glnx_shutil_mkdir_p_at (root_dfd, "tmp", 01777,
+                               cancellable, error))
     goto out;
-  g_clear_object (&child);
+  if (fchmodat (root_dfd, "tmp", 01777, 0) == -1)
+    {
+      glnx_set_prefix_error_from_errno (error, "chmod: %s", "tmp");
+      goto out;
+    }
 
   if (!ostree_sysroot_ensure_initialized (target_sysroot, cancellable, error))
     goto out;
