@@ -22,6 +22,7 @@
 
 #include <stdlib.h>
 
+#include "ot-main.h"
 #include "ot-admin-builtins.h"
 #include "ot-admin-functions.h"
 #include "ostree.h"
@@ -32,21 +33,22 @@ static GOptionEntry options[] = {
 };
 
 gboolean
-ot_admin_builtin_undeploy (int argc, char **argv, OstreeSysroot *sysroot, GCancellable *cancellable, GError **error)
+ot_admin_builtin_undeploy (int argc, char **argv, GCancellable *cancellable, GError **error)
 {
   gboolean ret = FALSE;
   GOptionContext *context;
+  glnx_unref_object OstreeSysroot *sysroot = NULL;
   const char *deploy_index_str;
   int deploy_index;
-  gs_unref_ptrarray GPtrArray *current_deployments = NULL;
-  gs_unref_object OstreeDeployment *booted_deployment = NULL;
-  gs_unref_object OstreeDeployment *target_deployment = NULL;
+  g_autoptr(GPtrArray) current_deployments = NULL;
+  glnx_unref_object OstreeDeployment *booted_deployment = NULL;
+  glnx_unref_object OstreeDeployment *target_deployment = NULL;
 
   context = g_option_context_new ("INDEX - Delete deployment INDEX");
 
-  g_option_context_add_main_entries (context, options, NULL);
-
-  if (!g_option_context_parse (context, &argc, &argv, error))
+  if (!ostree_admin_option_context_parse (context, options, &argc, &argv,
+                                          OSTREE_ADMIN_BUILTIN_FLAG_SUPERUSER,
+                                          &sysroot, cancellable, error))
     goto out;
 
   if (argc < 2)
@@ -62,20 +64,10 @@ ot_admin_builtin_undeploy (int argc, char **argv, OstreeSysroot *sysroot, GCance
   deploy_index_str = argv[1];
   deploy_index = atoi (deploy_index_str);
 
-  if (deploy_index < 0)
-    {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
-                   "Invalid index %d", deploy_index);
-      goto out;
-    }
-  if (deploy_index >= current_deployments->len)
-    {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
-                   "Out of range index %d, expected < %d", deploy_index, current_deployments->len);
-      goto out;
-    }
-  
-  target_deployment = g_object_ref (current_deployments->pdata[deploy_index]);
+  target_deployment = ot_admin_get_indexed_deployment (sysroot, deploy_index, error);
+  if (!target_deployment)
+    goto out;
+
   if (target_deployment == ostree_sysroot_get_booted_deployment (sysroot))
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
